# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""reader"""

from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from typing import Any, NamedTuple, TypeAlias, TypeVar

import numpy as np
import rasterio
import rasterio.crs
from affine import Affine
from odc.geo import CRS
from rasterio.io import DatasetReader
from typing_extensions import override

from datacube.drivers._types import (
    FutureGeoRasterReader,
    FutureNdarray,
    GeoRasterReader,
    RasterShape,
    RasterWindow,
    ReaderDriver,
    ReaderDriverEntry,
)
from datacube.storage import BandInfo
from datacube.utils import (
    get_part_from_uri,
    uri_to_local_path,
)


class Overrides(NamedTuple):
    crs: CRS | None
    transform: Affine | None
    nodata: float | int | None


RioWindow: TypeAlias = tuple[tuple[int, int], tuple[int, int]]  # pylint: disable=invalid-name
T = TypeVar("T")


def pick(a: T | None, b: T | None) -> T | None:
    """Return first non-None value or None if all are None"""
    return b if a is None else a


def _is_netcdf(fmt: str) -> bool:
    return fmt == "NetCDF"


def _roi_to_window(roi: RasterWindow | None, shape: RasterShape) -> RioWindow | None:
    if roi is None:
        return None

    def s2t(s: slice, n: int) -> tuple[int, int]:
        _in = 0 if s.start is None else s.start
        _out = n if s.stop is None else s.stop

        if _in < 0:
            _in += n
        if _out < 0:
            _out += n

        return _in, _out

    s1, s2 = (s2t(s, n) for s, n in zip(roi, shape))
    return s1, s2


def _dc_crs(crs: rasterio.crs.CRS | None) -> CRS | None:  # pylint: disable=c-extension-no-member
    """Convert RIO version of CRS to datacube"""
    if crs is None:
        return None

    if not crs.is_valid:
        return None

    if crs.is_epsg_code:
        return CRS(f"epsg:{crs.to_epsg()}")
    return CRS(crs.wkt)


def _read(
    src: DatasetReader,
    bidx: int,
    window: RasterWindow | None,
    out_shape: RasterShape | None,
) -> np.ndarray:
    return src.read(bidx, window=_roi_to_window(window, src.shape), out_shape=out_shape)


def _rio_uri(band: BandInfo) -> str:
    """
    - file uris are converted to file names
       - if also netcdf wrap in NETCDF:"${filename}":${layer}
    - All other protocols go through unmodified
    """
    if band.uri_scheme == "file":
        fname = str(uri_to_local_path(band.uri))

        if _is_netcdf(band.format):
            fname = f'NETCDF:"{fname}":{band.layer}'

        return fname

    return band.uri


def _rio_band_idx(band: BandInfo, src: DatasetReader) -> int:
    if band.band is not None:
        return band.band

    if not _is_netcdf(band.format):
        return 1

    bidx = get_part_from_uri(band.uri)
    if bidx is not None:
        return bidx

    if src.count == 1:  # Single-slice netcdf file
        return 1

    raise DeprecationWarning(
        "Stacked netcdf without explicit time index is not supported anymore"
    )


class RIOReader(GeoRasterReader):
    def __init__(
        self,
        src: DatasetReader,
        band_idx: int,
        pool: ThreadPoolExecutor,
        overrides: Overrides = Overrides(None, None, None),
    ) -> None:
        transform = pick(overrides.transform, src.transform)
        if transform is not None and transform.is_identity:
            transform = None

        self._src = src
        self._crs = overrides.crs or _dc_crs(src.crs)
        self._transform = transform
        self._nodata = pick(overrides.nodata, src.nodatavals[band_idx - 1])
        self._band_idx = band_idx
        self._dtype = src.dtypes[band_idx - 1]
        self._pool = pool

    @property
    @override
    def crs(self) -> CRS | None:
        return self._crs

    @property
    @override
    def transform(self) -> Affine | None:
        return self._transform

    @property
    @override
    def dtype(self) -> np.dtype:
        return np.dtype(self._dtype)

    @property
    @override
    def shape(self) -> RasterShape:
        return self._src.shape

    @property
    @override
    def nodata(self) -> int | float | None:
        return self._nodata

    @override
    def read(
        self, window: RasterWindow | None = None, out_shape: RasterShape | None = None
    ) -> FutureNdarray:
        return self._pool.submit(_read, self._src, self._band_idx, window, out_shape)


def _compute_overrides(src: DatasetReader, bi: BandInfo) -> Overrides:
    """If dataset is missing nodata, crs or transform."""
    crs, transform, nodata = None, None, None

    if src.crs is None or not src.crs.is_valid:
        crs = bi.crs

    if src.transform.is_identity:
        transform = bi.transform

    if src.nodata is None:
        nodata = bi.nodata

    return Overrides(crs=crs, transform=transform, nodata=nodata)


def _rdr_open(band: BandInfo, ctx: Any, pool: ThreadPoolExecutor) -> RIOReader:
    """Open file pointed by BandInfo and return RIOReader instance.

    raises Exception on failure

    TODO: start using ctx for handle cache
    """
    normalised_uri = _rio_uri(band)
    src = rasterio.open(normalised_uri, "r")
    bidx = _rio_band_idx(band, src)

    return RIOReader(src, bidx, pool, _compute_overrides(src, band))


class RIORdrDriver(ReaderDriver):
    def __init__(self, pool: ThreadPoolExecutor, cfg: dict) -> None:
        self._pool = pool
        self._cfg = cfg

    @override
    def new_load_context(self, bands: Iterable[BandInfo], old_ctx: Any | None) -> Any:
        return None  # TODO: implement file handle cache with this

    @override
    def open(self, band: BandInfo, ctx: Any) -> FutureGeoRasterReader:
        return self._pool.submit(_rdr_open, band, ctx, self._pool)


class RDEntry(ReaderDriverEntry):
    PROTOCOLS = ["file", "http", "https", "s3", "ftp", "zip"]
    FORMATS = ["GeoTIFF", "NetCDF", "JPEG2000"]

    @property
    @override
    def protocols(self) -> list[str]:
        return RDEntry.PROTOCOLS

    @property
    @override
    def formats(self) -> list[str]:
        return RDEntry.FORMATS

    @override
    def supports(self, protocol: str, fmt: str) -> bool:
        # TODO: might need better support matrix structures

        if fmt == "NetCDF":
            return protocol == "file"

        return True

    @override
    def new_instance(self, cfg: dict) -> ReaderDriver:
        cfg = cfg.copy()
        pool = cfg.pop("pool", None)
        if pool is None:
            max_workers = cfg.pop("max_workers", 1)
            pool = ThreadPoolExecutor(max_workers=max_workers)
        elif not isinstance(pool, ThreadPoolExecutor):
            if not cfg.pop("allow_custom_pool", False):
                raise ValueError("External `pool` should be a `ThreadPoolExecutor`")

        return RIORdrDriver(pool, cfg)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

from copy import deepcopy

import pytest

from datacube.model import Product
from datacube.utils import InvalidDocException

only_mandatory_fields = {
    "name": "ls7_nbar",
    "description": "description",
    "metadata_type": "eo",
    "metadata": {"product_type": "test"},
}


@pytest.mark.parametrize(
    "valid_product_update",
    [
        {},
        {"storage": {"crs": "EPSG:3577"}},
        # With the optional properties
        {
            "measurements": [
                {"name": "band_70", "dtype": "int16", "nodata": -999, "units": "1"}
            ]
        },
    ],
)
def test_accepts_valid_docs(valid_product_update) -> None:
    doc = deepcopy(only_mandatory_fields)
    doc.update(valid_product_update)
    # Should have no errors.
    Product.validate(doc)  # type: ignore[attr-defined]


def test_incomplete_product_is_invalid() -> None:
    # Invalid: An empty doc.
    with pytest.raises(InvalidDocException):
        Product.validate({})  # type: ignore[attr-defined]


# Changes to the above dict that should render it invalid.
@pytest.mark.parametrize(
    "invalid_product_update",
    [
        # Mandatory
        {"name": None},
        # Should be an object
        {"storage": "s"},
        # Should be a string
        {"description": 123},
        # Unknown property
        {"asdf": "asdf"},
        # Name must have alphanumeric & underscores only.
        {"name": " whitespace "},
        {"name": "with-dashes"},
        # Mappings
        {"mappings": {}},
        {"mappings": ""},
    ],
)
def test_rejects_invalid_docs(invalid_product_update) -> None:
    mapping = deepcopy(only_mandatory_fields)
    mapping.update(invalid_product_update)
    with pytest.raises(InvalidDocException):
        Product.validate(mapping)  # type: ignore[attr-defined]


@pytest.mark.parametrize(
    "valid_product_measurement",
    [
        {"name": "1", "dtype": "int16", "units": "1", "nodata": -999},
        # With the optional properties
        {
            "name": "red",
            "nodata": -999,
            "units": "1",
            "dtype": "int16",
            # TODO: flags/spectral
        },
    ],
)
def test_accepts_valid_measurements(valid_product_measurement) -> None:
    mapping = deepcopy(only_mandatory_fields)
    mapping["measurements"] = [valid_product_measurement]
    # Should have no errors.
    Product.validate(mapping)  # type: ignore[attr-defined]


# Changes to the above dict that should render it invalid.
@pytest.mark.parametrize(
    "invalid_product_measurement",
    [
        # no name
        {"nodata": -999},
        # nodata must be numeric
        {"name": "red", "nodata": "-999"},
        # Limited dtype options
        {"name": "red", "dtype": "asdf"},
        {"name": "red", "dtype": "intt13"},
        {"name": "red", "dtype": 13},
        # Unknown property
        {"name": "red", "asdf": "asdf"},
    ],
)
def test_rejects_invalid_measurements(invalid_product_measurement) -> None:
    mapping = deepcopy(only_mandatory_fields)
    mapping["measurements"] = {"10": invalid_product_measurement}
    with pytest.raises(InvalidDocException):
        Product.validate(mapping)  # type: ignore[attr-defined]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Datacube metadata generation from STAC.
"""

from ._eo3converter import infer_dc_product, stac2ds

__all__ = (
    "infer_dc_product",
    "stac2ds",
)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import copy
from collections.abc import Mapping
from typing import Any, cast

from datacube.model import Measurement
from datacube.utils import import_function
from datacube.utils.documents import parse_yaml

from .catalog import Catalog
from .impl import (
    Transformation,
    VirtualProduct,
    VirtualProductException,
    from_validated_recipe,
    virtual_product_kind,
)
from .transformations import (
    ApplyMask,
    Expressions,
    MakeMask,
    Rename,
    Select,
    ToFloat,
    XarrayReduction,
    day,
    earliest_time,
    fiscal_year,
    month,
    week,
    year,
)
from .utils import reject_keys

__all__ = ["Measurement", "Transformation", "construct"]


class NameResolver:
    """Apply a mapping from name to callable objects in a recipe."""

    def __init__(self, lookup_table) -> None:
        self.lookup_table = lookup_table

    def clone(self) -> "NameResolver":
        """Safely copy the resolver in order to possibly extend it."""
        return NameResolver(copy.deepcopy(self.lookup_table))

    @staticmethod
    def _assert(cond, msg) -> None:
        if not cond:
            raise VirtualProductException(msg)

    def construct(self, **recipe) -> VirtualProduct:
        """Validate recipe and construct virtual product."""

        get = recipe.get

        def lookup(name, namespace=None, kind: str = "transformation"):
            if callable(name):
                return name

            if (
                namespace is not None
                and namespace in self.lookup_table
                and name in self.lookup_table[namespace]
            ):
                result = self.lookup_table[namespace][name]
            else:
                try:
                    result = import_function(name)
                except (ImportError, AttributeError):
                    msg = f"could not resolve {kind} {name} in {recipe}"
                    raise VirtualProductException(msg) from None

            self._assert(callable(result), f"{kind} not callable in {recipe}")

            return result

        kind = virtual_product_kind(recipe)

        if kind == "product":
            func_keys = ["fuse_func", "dataset_predicate"]
            return from_validated_recipe(
                {
                    key: value
                    if key not in func_keys
                    else lookup(value, kind="function")
                    for key, value in recipe.items()
                }
            )

        if kind == "transform":
            cls_name = recipe["transform"]
            input_product = cast(Mapping, get("input"))

            self._assert(
                input_product is not None, f"no input for transformation in {recipe}"
            )

            return from_validated_recipe(
                dict(
                    transform=lookup(cls_name, "transform"),
                    input=self.construct(**input_product),
                    **reject_keys(recipe, ["transform", "input"]),
                )
            )

        if kind == "collate":
            self._assert(
                len(recipe["collate"]) > 0, f"no children for collate in {recipe}"
            )

            return from_validated_recipe(
                dict(
                    collate=[self.construct(**child) for child in recipe["collate"]],
                    **reject_keys(recipe, ["collate"]),
                )
            )

        if kind == "juxtapose":
            self._assert(
                len(recipe["juxtapose"]) > 0, f"no children for juxtapose in {recipe}"
            )

            return from_validated_recipe(
                dict(
                    juxtapose=[
                        self.construct(**child) for child in recipe["juxtapose"]
                    ],
                    **reject_keys(recipe, ["juxtapose"]),
                )
            )

        if kind == "aggregate":
            cls_name = recipe["aggregate"]
            input_product = cast(Mapping, get("input"))
            group_by = get("group_by")

            self._assert(
                input_product is not None, f"no input for aggregate in {recipe}"
            )
            self._assert(group_by is not None, f"no group_by for aggregate in {recipe}")

            return from_validated_recipe(
                dict(
                    aggregate=lookup(cls_name, "aggregate"),
                    group_by=lookup(group_by, "aggregate/group_by", kind="group_by"),
                    input=self.construct(**input_product),
                    **reject_keys(recipe, ["aggregate", "input", "group_by"]),
                )
            )

        if kind == "reproject":
            input_product = cast(Mapping, get("input"))
            output_crs = recipe["reproject"].get("output_crs")
            resolution = recipe["reproject"].get("resolution")

            self._assert(
                input_product is not None, f"no input for reproject in {recipe}"
            )
            self._assert(
                output_crs is not None, f"no output_crs for reproject in {recipe}"
            )
            self._assert(
                resolution is not None, f"no resolution for reproject in {recipe}"
            )

            return from_validated_recipe(
                dict(
                    reproject=recipe["reproject"],
                    input=self.construct(**input_product),
                    **reject_keys(recipe, ["reproject", "input"]),
                )
            )

        raise VirtualProductException(
            f"could not understand virtual product recipe: {recipe}"
        )

    def register(self, namespace: str, name: str, callable_obj) -> None:
        """Register a callable to the name resolver."""
        if namespace not in self.lookup_table:
            self.lookup_table[namespace] = {}

        if name in self.lookup_table[namespace]:
            raise VirtualProductException(
                f"name {name} under {namespace} is already registered"
            )

        self.lookup_table[namespace][name] = callable_obj


DEFAULT_RESOLVER = NameResolver(
    {
        "transform": {
            "make_mask": MakeMask,
            "apply_mask": ApplyMask,
            "to_float": ToFloat,
            "rename": Rename,
            "select": Select,
            "expressions": Expressions,
        },
        "aggregate": {"xarray_reduction": XarrayReduction},
        "aggregate/group_by": {
            "year": year,
            "month": month,
            "week": week,
            "day": day,
            "earliest_time": earliest_time,
            "fiscal_year": fiscal_year,
        },
    }
)


def construct(
    name_resolver: NameResolver | None = None, **recipe: Mapping[str, Any]
) -> VirtualProduct:
    """
    Create a virtual product from a specification dictionary.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return name_resolver.construct(**recipe)


def construct_from_yaml(
    recipe: str, name_resolver: NameResolver | None = None
) -> VirtualProduct:
    """
    Create a virtual product from a yaml recipe.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return construct(name_resolver=name_resolver, **parse_yaml(recipe))


def catalog_from_yaml(
    catalog_body: str, name_resolver: NameResolver | None = None
) -> Catalog:
    """
    Load a catalog of virtual products from a yaml document.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return Catalog(name_resolver, parse_yaml(catalog_body))


def catalog_from_file(filename: str, name_resolver=None) -> Catalog:
    """
    Load a catalog of virtual products from a yaml file.
    """
    with open(filename) as fl:
        return catalog_from_yaml(fl.read(), name_resolver=name_resolver)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Access methods for indexing datasets & products.
"""

import logging

from datacube.cfg import ODCConfig, ODCEnvironment

from .abstract import AbstractIndex as Index

_LOG: logging.Logger = logging.getLogger(__name__)


def index_connect(
    config_env: ODCEnvironment | None = None,
    application_name: str | None = None,
    validate_connection: bool = True,
) -> Index:
    """
    Create a Data Cube Index (as per config)

    It contains all the required connection parameters, but doesn't actually
    check that the server is available.

    :param config_env: Config environment to use. (optional)
    :param application_name: A short, alphanumeric name to identify this application.
    :param validate_connection: Validate database connection and schema immediately
    :raises datacube.index.Exceptions.IndexSetupError:
    """
    from datacube.drivers import index_driver_by_name

    if config_env is None:
        config_env = ODCConfig()[None]

    driver_name = config_env.index_driver
    index_driver = index_driver_by_name(driver_name)
    # No need to check for missing index driver - already checked during config parsing.
    assert index_driver is not None
    return index_driver.connect_to_index(
        config_env,
        application_name=application_name,
        validate_connection=validate_connection,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from dateutil.tz import UTC

from datacube.model import Range
from datacube.ui import parse_expressions


def test_parse_empty_str() -> None:
    q = parse_expressions("")
    assert q == {}


def test_id_search() -> None:
    q = parse_expressions("id = 26931d17-7a4e-4b55-98e7-d6777fb61df6")
    assert q["id"] == "26931d17-7a4e-4b55-98e7-d6777fb61df6"

    q = parse_expressions('id = "26931d17-7a4e-4b55-98e7-d6777fb61df6"')
    assert q["id"] == "26931d17-7a4e-4b55-98e7-d6777fb61df6"


def test_simple_string() -> None:
    q = parse_expressions("region_code = 56KKD", "i=10", "f=10.3")
    assert q["region_code"] == "56KKD"
    assert q["i"] == 10
    assert q["f"] == 10.3


def test_between_expression() -> None:
    q = parse_expressions("time in [2014, 2015]")
    assert "time" in q
    r = q["time"]
    assert isinstance(r, Range)
    assert isinstance(r.begin, datetime)
    assert isinstance(r.end, datetime)

    for k in ("lon", "lat", "x", "y"):
        q = parse_expressions(f"{k} in [10, 11.3]")
        assert k in q
        r = q[k]
        assert isinstance(r, Range)
        assert isinstance(r.begin, int | float)
        assert isinstance(r.end, int | float)
        assert r == Range(10, 11.3)


def test_parse_simple_expression() -> None:
    assert parse_expressions("platform = 4") == {"platform": 4}
    assert parse_expressions('platform = "LANDSAT_8"') == {"platform": "LANDSAT_8"}
    assert parse_expressions("platform = LANDSAT_8") == {"platform": "LANDSAT_8"}
    assert parse_expressions('platform = "LAND SAT_8"') == {"platform": "LAND SAT_8"}

    assert parse_expressions("lat in [4, 6]") == {"lat": Range(4, 6)}


def test_parse_uri_expression() -> None:
    assert parse_expressions("uri = file:///f/data/test.nc") == {
        "uri": "file:///f/data/test.nc"
    }
    assert parse_expressions('uri = "file:///f/data/test.nc"') == {
        "uri": "file:///f/data/test.nc"
    }
    assert parse_expressions('uri = "file:///f/data/test me.nc"') == {
        "uri": "file:///f/data/test me.nc"
    }
    assert parse_expressions("uri = file:///C:/f/data/test.nc") == {
        "uri": "file:///C:/f/data/test.nc"
    }
    assert parse_expressions('uri = "file:///C:/f/data/test.nc"') == {
        "uri": "file:///C:/f/data/test.nc"
    }
    assert parse_expressions('uri = "file:///C:/f/data/test me.nc"') == {
        "uri": "file:///C:/f/data/test me.nc"
    }


def test_parse_dates() -> None:
    assert parse_expressions("time in 2014-03-02") == {
        "time": Range(
            begin=datetime(2014, 3, 2, 0, 0, tzinfo=UTC),
            end=datetime(2014, 3, 2, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }

    assert parse_expressions("time in 2014-3-2") == {
        "time": Range(
            begin=datetime(2014, 3, 2, 0, 0, tzinfo=UTC),
            end=datetime(2014, 3, 2, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }

    # A missing day defaults to the first of the month.
    # They are probably better off using in-expessions in these cases (eg. "time in 2013-01"), but it's here
    # for backwards compatibility.
    march_2014 = {
        "time": Range(
            begin=datetime(2014, 3, 1, 0, 0, tzinfo=UTC),
            end=datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in 2014-03") == march_2014
    assert parse_expressions("time in 2014-3") == march_2014

    implied_feb_march_2014 = {
        "time": Range(
            begin=datetime(2014, 2, 1, 0, 0, tzinfo=UTC),
            end=datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in [2014-02, 2014-03]") == implied_feb_march_2014


def test_parse_date_ranges() -> None:
    eighth_march_2014 = {
        "time": Range(
            datetime(2014, 3, 8, tzinfo=UTC),
            datetime(2014, 3, 8, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in 2014-03-08") == eighth_march_2014
    assert parse_expressions("time in 2014-03-8") == eighth_march_2014

    march_2014 = {
        "time": Range(
            datetime(2014, 3, 1, tzinfo=UTC),
            datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in 2014-03") == march_2014
    assert parse_expressions("time in 2014-3") == march_2014
    # Leap year, 28 days
    feb_2014 = {
        "time": Range(
            datetime(2014, 2, 1, tzinfo=UTC),
            datetime(2014, 2, 28, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in 2014-02") == feb_2014
    assert parse_expressions("time in 2014-2") == feb_2014

    # Entire year
    year_2014 = {
        "time": Range(
            datetime(2014, 1, 1, tzinfo=UTC),
            datetime(2014, 12, 31, 23, 59, 59, 999999, tzinfo=UTC),
        )
    }
    assert parse_expressions("time in 2014") == year_2014


def test_parse_multiple_simple_expressions() -> None:
    # Multiple expressions in one command-line statement.
    # Mixed whitespace:
    between_exp = parse_expressions('platform=LS8 lat in [-4, 23.5] instrument="OTHER"')
    assert between_exp == {
        "platform": "LS8",
        "lat": Range(-4, 23.5),
        "instrument": "OTHER",
    }
    # Range(x,y) is "equal" to (x, y). Check explicitly that it's a range:
    assert between_exp["lat"].begin == -4



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from types import SimpleNamespace

import numpy as np
import pytest
from xarray import Dataset

from datacube import Datacube
from datacube.api.query import query_group_by
from datacube.testutils import (
    gen_tiff_dataset,
    mk_sample_dataset,
    mk_test_image,
)
from datacube.testutils.io import (
    get_raster_info,
    native_geobox,
    native_load,
    rio_slurp,
    rio_slurp_xarray,
    write_gtiff,
)
from datacube.testutils.iodriver import NetCDF
from datacube.utils import ignore_exceptions_if


def test_load_data(tmpdir) -> None:
    tmpdir = Path(str(tmpdir))

    group_by = query_group_by("time")
    spatial = {
        "resolution": (15, -15),
        "offset": (11230, 1381110),
    }

    nodata = -999
    aa = mk_test_image(96, 64, "int16", nodata=nodata)

    ds, geobox = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmpdir,
        prefix="ds1-",
        timestamp="2018-07-19",
        **spatial,
    )
    assert ds.time is not None

    ds2, _ = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmpdir,
        prefix="ds2-",
        timestamp="2018-07-19",
        **spatial,
    )
    assert ds.time is not None
    assert ds.time == ds2.time

    # FIXME: GroupBy expected.
    sources = Datacube.group_datasets([ds], "time")  # type: ignore[arg-type]
    sources2 = Datacube.group_datasets([ds, ds2], group_by)

    mm = ["aa"]
    mm = [ds.product.measurements[k] for k in mm]

    ds_data = Datacube.load_data(sources, geobox, mm)
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])

    custom_fuser_call_count = 0

    def custom_fuser(dest, delta) -> None:
        nonlocal custom_fuser_call_count
        custom_fuser_call_count += 1
        dest[:] += delta

    progress_call_data = []

    def progress_cbk(n, nt) -> None:
        progress_call_data.append((n, nt))

    ds_data = Datacube.load_data(
        sources2, geobox, mm, fuse_func=custom_fuser, progress_cbk=progress_cbk
    )
    assert ds_data.aa.nodata == nodata
    assert custom_fuser_call_count > 0
    np.testing.assert_array_equal(nodata + aa + aa, ds_data.aa.values[0])

    assert progress_call_data == [(1, 2), (2, 2)]

    ds_data = Datacube.load_data(sources, geobox, mm, dask_chunks={"x": 8, "y": 8})
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])


def test_load_data_dask(tmp_path) -> None:
    spatial = {
        "resolution": (15, -15),
        "offset": (11230, 1381110),
    }

    nodata = -999
    aa = mk_test_image(128, 128, "int16", nodata=nodata)

    ds, geobox = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmp_path,
        prefix="ds1-",
        timestamp="2018-07-19",
        **spatial,
    )

    # FIXME: GroupBy expected
    sources = Datacube.group_datasets([ds], "time")  # type: ignore[arg-type]
    mm = ds.product.measurements

    import dask

    dask.config.set(scheduler="synchronous")

    ds_data = Datacube.load_data(
        sources, geobox, mm, dask_chunks={"x": 50, "y": 67}
    )  # spatial dims not equal!
    ds_data.compute()
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])

    # Include an empty value area outside the data area, large enough to include completely empty chunks
    geobox = geobox.pad(100, 100)
    ds_data = Datacube.load_data(
        sources, geobox, mm, dask_chunks={"x": 50, "y": 67}
    )  # spatial dims not equal!

    ds_data.compute()
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0, 100:228, 100:228])


def test_load_data_with_url_mangling(tmpdir) -> None:
    actual_tmpdir = Path(str(tmpdir))
    recorded_tmpdir = Path(str(tmpdir / "not" / "actual" / "location"))

    def url_mangler(raw):
        actual_uri_root = actual_tmpdir.absolute().as_uri()
        recorded_uri_root = recorded_tmpdir.absolute().as_uri()
        return raw.replace(recorded_uri_root, actual_uri_root)

    group_by = query_group_by("time")
    spatial = {
        "resolution": (15, -15),
        "offset": (11230, 1381110),
    }

    nodata = -999
    aa = mk_test_image(96, 64, "int16", nodata=nodata)

    ds, geobox = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmpdir,
        prefix="ds1-",
        timestamp="2018-07-19",
        base_folder_of_record=recorded_tmpdir,
        **spatial,
    )
    assert ds.time is not None

    ds2, _ = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmpdir,
        prefix="ds2-",
        timestamp="2018-07-19",
        base_folder_of_record=recorded_tmpdir,
        **spatial,
    )
    assert ds.time is not None
    assert ds.time == ds2.time

    # FIXME: GroupBy expected.
    sources = Datacube.group_datasets([ds], "time")  # type: ignore[arg-type]
    sources2 = Datacube.group_datasets([ds, ds2], group_by)

    mm = ["aa"]
    mm = [ds.product.measurements[k] for k in mm]

    ds_data = Datacube.load_data(sources, geobox, mm, patch_url=url_mangler)
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])

    ds_data = Datacube.load_data(
        sources, geobox, mm, dask_chunks={"x": 8, "y": 8}, patch_url=url_mangler
    )
    assert ds_data.aa.nodata == nodata
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])

    custom_fuser_call_count = 0

    def custom_fuser(dest, delta) -> None:
        nonlocal custom_fuser_call_count
        custom_fuser_call_count += 1
        dest[:] += delta

    progress_call_data = []

    def progress_cbk(n, nt) -> None:
        progress_call_data.append((n, nt))

    ds_data = Datacube.load_data(
        sources2,
        geobox,
        mm,
        fuse_func=custom_fuser,
        progress_cbk=progress_cbk,
        patch_url=url_mangler,
    )
    assert ds_data.aa.nodata == nodata
    assert custom_fuser_call_count > 0
    np.testing.assert_array_equal(nodata + aa + aa, ds_data.aa.values[0])

    assert progress_call_data == [(1, 2), (2, 2)]


def test_load_data_cbk(tmpdir) -> None:
    from datacube.api import TerminateCurrentLoad

    tmpdir = Path(str(tmpdir))

    spatial = {
        "resolution": (15, -15),
        "offset": (11230, 1381110),
    }

    nodata = -999
    aa = mk_test_image(96, 64, "int16", nodata=nodata)

    bands = [
        SimpleNamespace(name=name, values=aa, nodata=nodata) for name in ["aa", "bb"]
    ]

    ds, geobox = gen_tiff_dataset(
        bands, tmpdir, prefix="ds1-", timestamp="2018-07-19", **spatial
    )
    assert ds.time is not None

    ds2, _ = gen_tiff_dataset(
        bands, tmpdir, prefix="ds2-", timestamp="2018-07-19", **spatial
    )
    assert ds.time is not None
    assert ds.time == ds2.time

    # FIXME: GroupBy expected.
    sources = Datacube.group_datasets([ds, ds2], "time")  # type: ignore[arg-type]
    progress_call_data = []

    def progress_cbk(n, nt) -> None:
        progress_call_data.append((n, nt))

    ds_data = Datacube.load_data(
        sources, geobox, ds.product.measurements, progress_cbk=progress_cbk
    )

    assert progress_call_data == [(1, 4), (2, 4), (3, 4), (4, 4)]
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])
    np.testing.assert_array_equal(aa, ds_data.bb.values[0])

    def progress_cbk_fail_early(n, nt):
        progress_call_data.append((n, nt))
        raise TerminateCurrentLoad()

    def progress_cbk_fail_early2(n, nt):
        progress_call_data.append((n, nt))
        if n > 1:
            raise KeyboardInterrupt()

    progress_call_data = []
    ds_data = Datacube.load_data(
        sources, geobox, ds.product.measurements, progress_cbk=progress_cbk_fail_early
    )

    assert progress_call_data == [(1, 4)]
    assert ds_data.dc_partial_load is True
    np.testing.assert_array_equal(aa, ds_data.aa.values[0])
    np.testing.assert_array_equal(nodata, ds_data.bb.values[0])

    progress_call_data = []
    ds_data = Datacube.load_data(
        sources, geobox, ds.product.measurements, progress_cbk=progress_cbk_fail_early2
    )

    assert ds_data.dc_partial_load is True
    assert progress_call_data == [(1, 4), (2, 4)]


def test_hdf5_lock_release_on_failure() -> None:
    from datacube.storage import BandInfo
    from datacube.storage._rio import HDF5_LOCK, RasterDatasetDataSource

    band = {"name": "xx", "layer": "xx", "dtype": "uint8", "units": "K", "nodata": 33}

    ds = mk_sample_dataset(
        [band],
        uri="file:///tmp/this_probably_doesnot_exist_37237827513/xx.nc",
        format=NetCDF,
    )
    src = RasterDatasetDataSource(BandInfo(ds, "xx"))

    with pytest.raises(OSError):  # noqa: SIM117
        with src.open():
            raise Exception("Did not expect to get here")

    assert not HDF5_LOCK._is_owned()  # type: ignore[attr-defined]


def test_rio_slurp(tmpdir) -> None:
    w, h, dtype, nodata, ndw = 96, 64, "int16", -999, 7

    pp = Path(str(tmpdir))

    aa = mk_test_image(w, h, dtype, nodata, nodata_width=ndw)

    assert aa.shape == (h, w)
    assert aa.dtype.name == dtype
    assert aa[10, 30] == (30 << 8) | 10
    assert aa[10, 11] == nodata

    aa0 = aa.copy()
    mm0 = write_gtiff(pp / "rio-slurp-aa.tif", aa, nodata=-999, overwrite=True)
    mm00 = write_gtiff(
        pp / "rio-slurp-aa-missing-nodata.tif", aa, nodata=None, overwrite=True
    )

    aa, mm = rio_slurp(mm0.path)
    np.testing.assert_array_equal(aa, aa0)
    assert mm.geobox == mm0.geobox
    assert aa.shape == mm.geobox.shape
    xx = rio_slurp_xarray(mm0.path)
    assert mm.geobox == xx.odc.geobox
    np.testing.assert_array_equal(xx.values, aa0)

    aa, mm = rio_slurp(mm0.path, aa0.shape)
    np.testing.assert_array_equal(aa, aa0)
    assert aa.shape == mm.geobox.shape
    assert mm.geobox is mm.src_geobox
    xx = rio_slurp_xarray(mm0.path, aa0.shape)
    assert mm.geobox == xx.odc.geobox
    np.testing.assert_array_equal(xx.values, aa0)

    aa, mm = rio_slurp(mm0.path, (3, 7))
    assert aa.shape == (3, 7)
    assert aa.shape == mm.geobox.shape
    assert mm.geobox != mm.src_geobox
    assert mm.src_geobox == mm0.geobox
    assert mm.geobox.extent == mm0.geobox.extent

    aa, mm = rio_slurp(mm0.path, aa0.shape)
    np.testing.assert_array_equal(aa, aa0)
    assert aa.shape == mm.geobox.shape

    aa, mm = rio_slurp(mm0.path, mm0.geobox, resampling="nearest")
    np.testing.assert_array_equal(aa, aa0)
    xx = rio_slurp_xarray(mm0.path, mm0.geobox)
    assert mm.geobox == xx.odc.geobox
    np.testing.assert_array_equal(xx.values, aa0)

    aa, mm = rio_slurp(mm0.path, geobox=mm0.geobox, dtype="float32")
    assert aa.dtype == "float32"
    np.testing.assert_array_equal(aa, aa0.astype("float32"))
    xx = rio_slurp_xarray(mm0.path, geobox=mm0.geobox)
    assert mm.geobox == xx.odc.geobox
    assert mm.nodata == xx.nodata
    np.testing.assert_array_equal(xx.values, aa0)

    aa, mm = rio_slurp(mm0.path, mm0.geobox, dst_nodata=-33)
    np.testing.assert_array_equal(aa == -33, aa0 == -999)

    aa, mm = rio_slurp(mm00.path, mm00.geobox, dst_nodata=None)
    np.testing.assert_array_equal(aa, aa0)


def test_rio_slurp_with_geobox(tmpdir) -> None:
    w, h, dtype, nodata, ndw = 96, 64, "int16", -999, 7

    pp = Path(str(tmpdir))
    aa = mk_test_image(w, h, dtype, nodata, nodata_width=ndw)
    assert aa.dtype.name == dtype
    assert aa[10, 30] == (30 << 8) | 10
    assert aa[10, 11] == nodata

    aa = np.stack([aa, aa[::-1, ::-1]])
    assert aa.shape == (2, h, w)
    aa0 = aa.copy()

    mm = write_gtiff(pp / "rio-slurp-aa.tif", aa, nodata=-999, overwrite=True)
    assert mm.count == 2

    aa, mm = rio_slurp(mm.path, mm.geobox)
    assert aa.shape == aa0.shape
    np.testing.assert_array_equal(aa, aa0)


def test_missing_file_handling() -> None:
    with pytest.raises(IOError):
        rio_slurp("no-such-file.tiff")

    # by default should catch any exception
    with ignore_exceptions_if(True):
        rio_slurp("no-such-file.tiff")

    # this is equivalent to previous default behaviour, note that missing http
    # resources are not OSError
    with ignore_exceptions_if(True, (OSError,)):
        rio_slurp("no-such-file.tiff")

    # check that only requested exceptions are caught
    with pytest.raises(IOError):  # noqa: SIM117
        with ignore_exceptions_if(True, (ValueError, ArithmeticError)):
            rio_slurp("no-such-file.tiff")


def test_native_load(tmpdir) -> None:
    tmpdir = Path(str(tmpdir))
    spatial = {
        "resolution": (15, -15),
        "offset": (11230, 1381110),
    }
    nodata = -999
    aa = mk_test_image(96, 64, "int16", nodata=nodata)
    cc = mk_test_image(32, 16, "int16", nodata=nodata)

    bands = [
        SimpleNamespace(name=name, values=aa, nodata=nodata) for name in ["aa", "bb"]
    ]
    bands.append(SimpleNamespace(name="cc", values=cc, nodata=nodata))

    ds, geobox = gen_tiff_dataset(
        bands[:2], tmpdir, prefix="ds1-", timestamp="2018-07-19", **spatial
    )

    assert set(get_raster_info(ds)) == set(ds.measurements)

    xx = native_load([ds], ["aa", "bb"], Datacube.group_datasets, "time")
    assert isinstance(xx, Dataset)
    assert xx.odc.geobox == geobox
    np.testing.assert_array_equal(aa, xx.isel(time=0).aa.values)
    np.testing.assert_array_equal(aa, xx.isel(time=0).bb.values)

    ds, geobox_cc = gen_tiff_dataset(
        bands, tmpdir, prefix="ds2-", timestamp="2018-07-19", **spatial
    )

    # cc is different size from aa,bb
    # cc is reprojected
    xx = native_load([ds], ["aa", "bb", "cc"], Datacube.group_datasets, "time")
    assert isinstance(xx, Dataset)
    assert xx.odc.geobox == geobox
    assert xx.odc.geobox != geobox_cc
    np.testing.assert_array_equal(aa, xx.isel(time=0).aa.values)
    np.testing.assert_array_equal(aa, xx.isel(time=0).bb.values)

    assert native_geobox(ds, basis="aa") == geobox
    xx = native_load(
        [ds], ["aa", "bb", "cc"], Datacube.group_datasets, "time", basis="aa"
    )
    assert isinstance(xx, Dataset)
    assert xx.odc.geobox == geobox
    np.testing.assert_array_equal(aa, xx.isel(time=0).aa.values)
    np.testing.assert_array_equal(aa, xx.isel(time=0).bb.values)

    # cc is compatible with self
    xx = native_load([ds], ["cc"], Datacube.group_datasets, "time")
    assert isinstance(xx, Dataset)
    assert xx.odc.geobox == geobox_cc
    np.testing.assert_array_equal(cc, xx.isel(time=0).cc.values)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

"""
Module for JSON type aliases.
"""

from __future__ import annotations

from typing import TypeAlias

JsonAtom: TypeAlias = None | bool | str | float | int
JsonLike: TypeAlias = JsonAtom | list["JsonLike"] | dict[str, "JsonLike"]
JsonDict: TypeAlias = dict[str, JsonLike]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock
from uuid import UUID

import pytest

from datacube import Datacube
from datacube.cfg import ODCEnvironment
from datacube.testutils import suppress_deprecations

test_uuid = UUID("4ec8fe97-e8b9-11e4-87ff-1040f381a756")


def empty(iterable) -> bool:
    for _ in iterable:
        return False
    return True


def test_init_null(null_config: ODCEnvironment) -> None:
    from datacube.drivers.indexes import index_cache

    idxs = index_cache()
    assert "default" in idxs._drivers
    assert "null" in idxs._drivers
    with Datacube(env=null_config, validate_connection=True) as dc:
        assert dc.index.url == "null"
        assert dc.index.environment.index_driver == "null"


def test_null_user_resource(null_config: ODCEnvironment) -> None:
    with Datacube(env=null_config, validate_connection=True) as dc:
        assert empty(dc.index.users.list_users())
        with pytest.raises(NotImplementedError):
            dc.index.users.create_user("user1", "password2", "role1")
        with pytest.raises(NotImplementedError):
            dc.index.users.delete_user("user1", "user2")
        with pytest.raises(NotImplementedError):
            dc.index.users.grant_role("role1", "user1", "user2")


def test_null_metadata_types_resource(null_config: ODCEnvironment) -> None:
    with Datacube(env=null_config, validate_connection=True) as dc:
        assert dc.index.metadata_types.get_all() == []
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.from_doc({})
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.add(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.can_update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.update_document({})
        with pytest.raises(KeyError):
            dc.index.metadata_types.get_unsafe(1)
        with pytest.raises(KeyError):
            dc.index.metadata_types.get_by_name_unsafe("eo")
        with pytest.raises(NotImplementedError):
            dc.index.metadata_types.check_field_indexes()


def test_null_product_resource(null_config: ODCEnvironment) -> None:
    with Datacube(env=null_config, validate_connection=True) as dc:
        assert dc.index.products.get_all() == []
        assert dc.index.products.search_robust(foo="bar", baz=12) == []
        assert empty(dc.index.products.get_with_fields(["foo", "bar"]))
        assert empty(dc.index.products.get_field_names())
        with pytest.raises(KeyError):
            dc.index.products.spatial_extent("a_prod")
        with pytest.raises(KeyError):
            dc.index.products.temporal_extent("a_prod")
        with pytest.raises(KeyError):
            dc.index.products.get_unsafe(1)
        with pytest.raises(KeyError):
            dc.index.products.get_by_name_unsafe("product1")
        with pytest.raises(NotImplementedError):
            dc.index.products.add(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.products.can_update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.products.update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.products.delete(MagicMock())


def test_null_dataset_resource(null_config: ODCEnvironment) -> None:
    with Datacube(env=null_config, validate_connection=True) as dc:
        assert dc.index.datasets.get(test_uuid) is None
        assert dc.index.datasets.bulk_get([test_uuid, "foo"]) == []
        assert dc.index.datasets.get_derived(test_uuid) == []
        assert not dc.index.datasets.has(test_uuid)
        assert dc.index.datasets.bulk_has([test_uuid, "foo"]) == [False, False]
        with pytest.raises(NotImplementedError):
            dc.index.datasets.add(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.datasets.can_update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.datasets.update(MagicMock())
        with pytest.raises(NotImplementedError):
            dc.index.datasets.archive([test_uuid, "foo"])
        with pytest.raises(NotImplementedError):
            dc.index.datasets.restore([test_uuid, "foo"])
        with pytest.raises(NotImplementedError):
            dc.index.datasets.purge([test_uuid, "foo"])

        assert empty(dc.index.datasets.get_all_dataset_ids(True))
        assert dc.index.datasets.get_location(test_uuid) is None
        assert dc.index.datasets.get_datasets_for_location("http://a.uri/test") == []

        with suppress_deprecations():
            assert empty(
                dc.index.datasets.get_field_names()
            )  # DEPRECATED WRAPPER METHOD
            assert (
                dc.index.datasets.get_locations(test_uuid) == []
            )  # Test deprecated method
            assert (
                dc.index.datasets.get_archived_locations(test_uuid) == []
            )  # Test deprecated method
            assert (
                dc.index.datasets.get_archived_location_times(test_uuid) == []
            )  # Test deprecated method

            with pytest.raises(NotImplementedError):
                dc.index.datasets.add_location(
                    test_uuid, "http://a.uri/test"
                )  # Test deprecated method
            with pytest.raises(NotImplementedError):
                dc.index.datasets.remove_location(
                    test_uuid, "http://a.uri/test"
                )  # Test deprecated method
            with pytest.raises(NotImplementedError):
                dc.index.datasets.archive_location(
                    test_uuid, "http://a.uri/test"
                )  # Test deprecated method
            with pytest.raises(NotImplementedError):
                dc.index.datasets.restore_location(
                    test_uuid, "http://a.uri/test"
                )  # Test deprecated method
        with pytest.raises(KeyError):
            dc.index.datasets.temporal_extent(ids=[test_uuid])
        assert dc.index.datasets.spatial_extent(ids=[test_uuid]) is None
        with suppress_deprecations(), pytest.raises(KeyError):
            dc.index.datasets.get_product_time_bounds(
                "product1"
            )  # Test deprecated method

        assert dc.index.datasets.search_product_duplicates(MagicMock()) == []
        assert dc.index.datasets.search_by_metadata({}) == []
        assert dc.index.datasets.search(foo="bar", baz=12) == []
        assert dc.index.datasets.search_by_product(foo="bar", baz=12) == []
        assert (
            list(dc.index.datasets.search_returning(["foo", "bar"], foo="bar", baz=12))
            == []
        )
        assert dc.index.datasets.count(foo="bar", baz=12) == 0
        assert dc.index.datasets.count_by_product(foo="bar", baz=12) == []
        assert (
            dc.index.datasets.count_by_product_through_time(
                "1 month", foo="bar", baz=12
            )
            == []
        )
        assert (
            dc.index.datasets.count_product_through_time("1 month", foo="bar", baz=12)
            == []
        )
        with suppress_deprecations():
            # Coverage test of deprecated method
            assert dc.index.datasets.search_summaries(foo="bar", baz=12) == []
            # Coverage test of deprecated base class method
            assert dc.index.datasets.search_eager(foo="bar", baz=12) == []
        assert (
            dc.index.datasets.search_returning_datasets_light(
                ("foo", "baz"), foo="bar", baz=12
            )
            == []
        )


def test_null_transactions(null_config: ODCEnvironment) -> None:
    with Datacube(env=null_config, validate_connection=True) as dc:
        trans = dc.index.transaction()
        assert not trans.active
        trans.begin()
        assert trans.active
        trans.commit()
        assert not trans.active
        trans.begin()
        assert dc.index.thread_transaction() == trans
        with pytest.raises(ValueError):
            trans.begin()
        trans.rollback()
        assert not trans.active
        assert dc.index.thread_transaction() is None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Modules for interfacing with the index/database.
"""

from datacube.index.abstract import AbstractIndex as Index

from ._api import index_connect
from ._spatial import extract_geom_from_query, strip_all_spatial_fields_from_query
from .exceptions import DuplicateRecordError, IndexSetupError, MissingRecordError
from .fields import UnknownFieldError

__all__ = [
    "DuplicateRecordError",
    "Index",
    "IndexSetupError",
    "MissingRecordError",
    "UnknownFieldError",
    "extract_geom_from_query",
    "index_connect",
    "strip_all_spatial_fields_from_query",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Datacube command-line interface
"""

import datacube.scripts.dataset
import datacube.scripts.metadata
import datacube.scripts.product
import datacube.scripts.spindex
import datacube.scripts.system
import datacube.scripts.user  # noqa: F401
from datacube.ui.click import cli

if __name__ == "__main__":
    cli()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import math
from pathlib import Path
from types import SimpleNamespace

import numpy as np
import pytest
import xarray as xr
from dask.base import is_dask_collection
from dask.delayed import Delayed

from datacube import Datacube
from datacube.testutils import gen_tiff_dataset, mk_test_image, suppress_deprecations
from datacube.testutils.io import native_load, rio_slurp, rio_slurp_xarray
from datacube.utils.cog import _write_cog, to_cog, write_cog


def gen_test_data(prefix, dask=False, shape=None, dtype="int16", nodata=-999):
    w, h, ndw = 96, 64, 7
    if shape is not None:
        h, w = shape

    aa = mk_test_image(w, h, dtype, nodata, nodata_width=ndw)

    ds, gbox = gen_tiff_dataset(
        SimpleNamespace(name="aa", values=aa, nodata=nodata), prefix
    )
    extras = {}

    if dask:
        extras.update(dask_chunks={"time": 1})

    xx = native_load([ds], ["aa"], Datacube.group_datasets, "time", **extras)

    return xx.aa.isel(time=0), ds


@pytest.mark.parametrize(
    "opts",
    [
        {},
        {"use_windowed_writes": True},
        {
            "intermediate_compression": {"compress": "deflate", "zlevel": 1},
            "use_windowed_writes": True,
        },
        {"intermediate_compression": True},
        {"intermediate_compression": "deflate"},
    ],
)
def test_cog_file(tmpdir, opts) -> None:
    pp = Path(str(tmpdir))
    xx, ds = gen_test_data(pp)

    with suppress_deprecations():
        # write to file
        ff = write_cog(  # Coverage test of deprecated function.
            xx, pp / "cog.tif", **opts
        )
    assert isinstance(ff, Path)
    assert ff == pp / "cog.tif"
    assert ff.exists()

    yy = rio_slurp_xarray(pp / "cog.tif")
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata

    with suppress_deprecations():
        _write_cog(  # Test of deprecated function
            np.stack([xx.values, xx.values]),
            xx.odc.geobox,
            pp / "cog-2-bands.tif",
            overview_levels=[],
            **opts,
        )

    yy, mm = rio_slurp(pp / "cog-2-bands.tif")
    assert mm.geobox == xx.odc.geobox
    assert yy.shape == (2, *xx.shape)
    np.testing.assert_array_equal(yy[0], xx.values)
    np.testing.assert_array_equal(yy[1], xx.values)

    with (
        pytest.raises(ValueError, match="Need 2d or 3d ndarray on input"),
        suppress_deprecations(),
    ):
        _write_cog(
            xx.values.ravel(), xx.odc.geobox, pp / "wontwrite.tif"
        )  # Test of deprecated function

    # sizes that are not multiples of 16
    # also check that supplying `nodata=` doesn't break things
    xx_odd = xx[:23, :63]
    with suppress_deprecations():
        ff = write_cog(  # Coverage test of deprecated function
            xx_odd, pp / "cog_odd.tif", nodata=xx_odd.attrs["nodata"], **opts
        )
    assert isinstance(ff, Path)
    assert ff == pp / "cog_odd.tif"
    assert ff.exists()

    yy = rio_slurp_xarray(pp / "cog_odd.tif")
    np.testing.assert_array_equal(yy.values, xx_odd.values)
    assert yy.odc.geobox == xx_odd.odc.geobox
    assert yy.nodata == xx_odd.nodata

    with suppress_deprecations(), pytest.warns(UserWarning):
        write_cog(
            xx, pp / "cog_badblocksize.tif", blocksize=50
        )  # Test of deprecated method

    # check writing floating point COG with no explicit nodata
    zz, ds = gen_test_data(pp, dtype="float32", nodata=None)
    # write to file
    with suppress_deprecations():
        ff = write_cog(zz, pp / "cog_float.tif", **opts)
    assert isinstance(ff, Path)
    assert ff == pp / "cog_float.tif"
    assert ff.exists()
    aa = rio_slurp_xarray(pp / "cog_float.tif")
    assert aa.attrs["nodata"] == "nan" or math.isnan(aa.attrs["nodata"])


def test_cog_file_dask(tmpdir) -> None:
    pp = Path(str(tmpdir))
    xx, ds = gen_test_data(pp, dask=True)
    assert is_dask_collection(xx)

    path = pp / "cog.tif"
    with suppress_deprecations():
        ff = write_cog(xx, path, overview_levels=[2, 4])  # Test of deprecated method
        assert isinstance(ff, Delayed)
        assert path.exists() is False
        assert ff.compute() == path
    assert path.exists()

    yy = rio_slurp_xarray(pp / "cog.tif")
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata


@pytest.mark.parametrize("shape", [None, (1024, 512)])
def test_cog_mem(tmpdir, shape) -> None:
    pp = Path(str(tmpdir))
    xx, ds = gen_test_data(pp, shape=shape)

    # write to memory 1
    with suppress_deprecations():
        bb = write_cog(xx, ":mem:")  # Test of deprecated function
    assert isinstance(bb, bytes)
    path = pp / "cog1.tiff"
    with open(str(path), "wb") as f:
        f.write(bb)

    yy = rio_slurp_xarray(path)
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata

    # write to memory 2
    with suppress_deprecations():
        bb = to_cog(xx)  # Test of deprecated function
    assert isinstance(bb, bytes)
    path = pp / "cog2.tiff"
    with open(str(path), "wb") as f:
        f.write(bb)

    yy = rio_slurp_xarray(path)
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata

    # write to memory 3 -- no overviews
    with suppress_deprecations():
        bb = to_cog(xx, overview_levels=[])  # Test of deprecated function
    assert isinstance(bb, bytes)
    path = pp / "cog3.tiff"
    with open(str(path), "wb") as f:
        f.write(bb)

    yy = rio_slurp_xarray(path)
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata


def test_cog_mem_dask(tmpdir) -> None:
    pp = Path(str(tmpdir))
    xx, ds = gen_test_data(pp, dask=True)

    # write to memory 1
    with suppress_deprecations():
        bb = write_cog(xx, ":mem:")  # Test of deprecated method
        assert isinstance(bb, Delayed)
        bb = bb.compute()
        assert isinstance(bb, bytes)

    path = pp / "cog1.tiff"
    with open(str(path), "wb") as f:
        f.write(bb)

    yy = rio_slurp_xarray(path)
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata

    # write to memory 2
    with suppress_deprecations():
        bb = to_cog(xx)  # Test of deprecated function
        assert isinstance(bb, Delayed)
        bb = bb.compute()
        assert isinstance(bb, bytes)
    path = pp / "cog2.tiff"
    with open(str(path), "wb") as f:
        f.write(bb)

    yy = rio_slurp_xarray(path)
    np.testing.assert_array_equal(yy.values, xx.values)
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.nodata == xx.nodata


@pytest.mark.parametrize("use_windowed_writes", [False, True])
def test_cog_rgba(tmpdir, use_windowed_writes) -> None:
    pp = Path(str(tmpdir))
    xx, ds = gen_test_data(pp)
    pix = np.dstack([xx.values] * 4)
    rgba = xr.DataArray(pix, attrs=xx.attrs, dims=("y", "x", "band"), coords=xx.coords)
    assert rgba.odc.geobox == xx.odc.geobox
    assert rgba.shape[:2] == rgba.odc.geobox.shape

    with suppress_deprecations():
        ff = write_cog(
            rgba, pp / "cog.tif", use_windowed_writes=use_windowed_writes
        )  # Test of deprecated function
    yy = rio_slurp_xarray(ff)

    assert yy.odc.geobox == rgba.odc.geobox
    assert yy.shape == rgba.shape
    np.testing.assert_array_equal(yy.values, rgba.values)

    with pytest.raises(ValueError), suppress_deprecations():
        _write_cog(  # Test of deprecated function
            rgba.values[1:, :, :],
            rgba.odc.geobox,
            ":mem:",
            use_windowed_writes=use_windowed_writes,
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

import pytest

from datacube.drivers.postgres._fields import (
    IntDocField,
    NumericRangeDocField,
    RangeDocField,
    SimpleDocField,
    parse_fields,
)
from datacube.drivers.postgres._schema import DATASET
from datacube.model import Range
from datacube.utils.uris import split_uri


def _assert_same(obj1, obj2) -> None:
    assert obj1.__class__ == obj2.__class__
    assert obj1.__dict__ == obj2.__dict__


def test_split_uri() -> None:
    assert split_uri("http://test.com/something.txt") == (
        "http",
        "//test.com/something.txt",
    )
    assert split_uri("eods:LS7_ETM_SYS_P31_GALPGS01-002_101_065_20160127") == (
        "eods",
        "LS7_ETM_SYS_P31_GALPGS01-002_101_065_20160127",
    )
    assert split_uri(
        "file://rhe-test-dev.prod.lan/data/fromASA/LANDSAT-7.89274.S4A2C1D3R3"
    ) == ("file", "//rhe-test-dev.prod.lan/data/fromASA/LANDSAT-7.89274.S4A2C1D3R3")
    assert split_uri("file:///C:/tmp/first/something.yaml") == (
        "file",
        "///C:/tmp/first/something.yaml",
    )

    with pytest.raises(ValueError):
        split_uri("/no/semicolon")


def test_get_single_field() -> None:
    fields = parse_fields(
        {
            "platform": {"description": "Satellite", "offset": ("platform", "code")},
            "instrument": {"offset": ("instrument", "name")},
        },
        DATASET.c.metadata,
    )
    assert set(fields.keys()) == {"platform", "instrument"}
    field = fields["platform"]
    _assert_same(
        field,
        SimpleDocField(
            "platform",
            "Satellite",
            DATASET.c.metadata,
            True,
            offset=("platform", "code"),
        ),
    )
    assert isinstance(field, SimpleDocField)
    assert field.extract({"platform": {"code": "turtle"}}) == "turtle"
    assert field.extract({"platform": {"code": None}}) is None
    assert field.extract({}) is None


def test_get_multi_field() -> None:
    fields = parse_fields(
        {
            "orbit": {
                "description": "Orbit number",
                "type": "integer",
                "offset": [("acquisition", "platform_orbit"), ("orbit",)],
            }
        },
        DATASET.c.metadata,
    )
    assert set(fields.keys()) == {"orbit"}

    field = fields["orbit"]
    _assert_same(
        field,
        IntDocField(
            "orbit",
            "Orbit number",
            DATASET.c.metadata,
            True,
            offset=[("acquisition", "platform_orbit"), ("orbit",)],
        ),
    )
    assert isinstance(field, SimpleDocField)
    assert field.extract({"platform": {"code": "turtle"}}) is None
    assert field.extract({"acquisition": {"platform_orbit": 5}}) == 5
    assert field.extract({"orbit": 10}) == 10
    # It chooses the first listed field with a non-null value
    assert field.extract({"orbit": 10, "acquisition": {"platform_orbit": 5}}) == 5


def test_get_range_field() -> None:
    storage_fields = parse_fields(
        {
            "lat": {
                "type": "float-range",
                "description": "latitude range",
                "max_offset": [["extents", "geospatial_lat_max"]],
                "min_offset": [
                    ["extents", "geospatial_lat_other"],
                    ["extents", "geospatial_lat_min"],
                ],
            },
        },
        DATASET.c.metadata,
    )
    field = storage_fields["lat"]
    _assert_same(
        field,
        NumericRangeDocField(
            "lat",
            "latitude range",
            DATASET.c.metadata,
            True,
            max_offset=[["extents", "geospatial_lat_max"]],
            min_offset=[
                ["extents", "geospatial_lat_other"],
                ["extents", "geospatial_lat_min"],
            ],
        ),
    )
    assert isinstance(field, RangeDocField)
    extracted = field.extract(
        {"extents": {"geospatial_lat_min": 2, "geospatial_lat_max": 4}}
    )
    assert extracted == Range(begin=2, end=4)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
from collections.abc import Iterable
from types import SimpleNamespace
from unittest.mock import MagicMock
from uuid import UUID

import numpy as np
import pytest
import xarray as xr

from datacube import Datacube
from datacube.api.core import _calculate_chunk_sizes, output_geobox
from datacube.api.query import GroupBy
from datacube.model import Dataset
from datacube.testutils import mk_sample_dataset, suppress_deprecations
from datacube.testutils.geom import AlbersGS


def test_grouping_datasets() -> None:
    def group_func(d):
        return d.time

    dimension = "time"
    units = None
    datasets = [
        SimpleNamespace(
            time=datetime.datetime(2016, 1, 1), value="foo", id=UUID(int=10)
        ),
        SimpleNamespace(
            time=datetime.datetime(2016, 2, 1), value="bar", id=UUID(int=1)
        ),
        SimpleNamespace(
            time=datetime.datetime(2016, 1, 1), value="flim", id=UUID(int=9)
        ),
    ]

    group_by = GroupBy(group_func, dimension, units, sort_key=group_func)
    # FIXME: Dataset expected.
    grouped = Datacube.group_datasets(datasets, group_by)  # type: ignore[arg-type]
    dss = grouped.isel(time=0).values[()]
    assert isinstance(dss, tuple)
    assert len(dss) == 2
    assert [ds.value for ds in dss] == ["flim", "foo"]

    dss = grouped.isel(time=1).values[()]
    assert isinstance(dss, tuple)
    assert len(dss) == 1
    assert [ds.value for ds in dss] == ["bar"]

    assert str(grouped.time.dtype) == "datetime64[ns]"
    # FIXME: Illegal slice index.
    assert grouped.loc["2016-01-01":"2016-01-15"]  # type: ignore[misc]


def test_group_datasets_by_time() -> None:
    bands = [{"name": "a"}]
    # Same time instant but one explicitly marked as UTC
    ds1 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00Z")
    ds2 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00")
    # Same "time" but in a different timezone, and actually later
    ds3 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00-1")
    assert ds1.center_time.tzinfo is not None
    assert ds2.center_time.tzinfo is None
    assert ds3.center_time.tzinfo is not None

    # FIXME: GroupBy expected.
    xx = Datacube.group_datasets([ds1, ds2, ds3], "time")  # type: ignore[arg-type]
    assert xx.time.shape == (2,)
    assert len(xx.data[0]) == 2
    assert len(xx.data[1]) == 1


def test_grouped_datasets_should_be_in_consistent_order() -> None:
    datasets = [
        {"time": datetime.datetime(2016, 1, 1, 0, 1), "value": "foo"},
        {"time": datetime.datetime(2016, 1, 1, 0, 2), "value": "flim"},
        {"time": datetime.datetime(2016, 2, 1, 0, 1), "value": "bar"},
    ]
    # FIXME: GroupBy expected.
    grouped = _group_datasets_by_date(datasets)  # type: ignore[arg-type]

    # Swap the two elements which get grouped together
    datasets[0], datasets[1] = datasets[1], datasets[0]
    grouped_2 = _group_datasets_by_date(datasets)  # type: ignore[arg-type]

    assert len(grouped) == len(grouped_2) == 2
    assert all(grouped.values == grouped_2.values)


def _group_datasets_by_date(datasets: Iterable[Dataset]) -> xr.DataArray:
    def group_func(d):
        return d["time"].date()

    def sort_key(d):
        return d["time"]

    dimension = "time"
    units = None

    group_by = GroupBy(group_func, dimension, units, sort_key)
    return Datacube.group_datasets(datasets, group_by)


def test_dask_chunks() -> None:
    coords = {"time": np.arange(10)}

    sources = xr.DataArray(coords["time"], coords=coords, dims=list(coords))
    geobox = AlbersGS.tile_geobox((0, 0))[:6, :7]

    assert geobox.dimensions == ("y", "x")
    assert sources.dims == ("time",)

    assert _calculate_chunk_sizes(sources, geobox, {}) == ((1,), (6, 7))
    assert _calculate_chunk_sizes(sources, geobox, {"time": -1}) == ((10,), (6, 7))
    assert _calculate_chunk_sizes(sources, geobox, {"time": "auto", "x": "auto"}) == (
        (1,),
        (6, 7),
    )
    assert _calculate_chunk_sizes(sources, geobox, {"y": -1, "x": 3}) == ((1,), (6, 3))
    assert _calculate_chunk_sizes(sources, geobox, {"y": 2, "x": 3}) == ((1,), (2, 3))

    with pytest.raises(ValueError):
        _calculate_chunk_sizes(sources, geobox, {"x": "aouto"})  # type: ignore[dict-item]

    with pytest.raises(KeyError):
        _calculate_chunk_sizes(sources, geobox, {"zz": 1})


def test_index_validation() -> None:
    index = MagicMock()
    with pytest.raises(ValueError) as e:
        Datacube(
            index=index,
            config=["/a/path", "/a/nother/path"],
            env="prod",
            app="this_is_me",
            raw_config="{}",
        )
    estr = str(e.value)
    assert "config,raw_config,app,env" in estr


def test_output_geobox() -> None:
    from odc.geo.crs import CRS as ODCGeoCRS  # noqa: N811
    from odc.geo.geobox import GeoBox as ODCGeoGeoBox

    with suppress_deprecations():
        from datacube.utils.geometry import CRS as LegacyCRS  # noqa: N811
        from datacube.utils.geometry import GeoBox as LegacyGeoGeoBox
    from odc.geo.xr import xr_zeros

    odc_gbox = ODCGeoGeoBox.from_bbox(
        (-2_000_000, -5_000_000, 2_250_000, -1_000_000), "epsg:3577", resolution=1000
    )
    legacy_gbox = LegacyGeoGeoBox(
        width=odc_gbox.width,
        height=odc_gbox.height,
        affine=odc_gbox.affine,
        crs=LegacyCRS(str(odc_gbox.crs)),
    )

    assert legacy_gbox != odc_gbox

    xra = xr_zeros(odc_gbox, chunks=(1000, 1000))

    assert xra.odc.geobox == odc_gbox

    gbox = output_geobox(like=xra)
    assert gbox == odc_gbox

    gbox = output_geobox(like=legacy_gbox)
    assert gbox == odc_gbox
    assert isinstance(gbox.crs, ODCGeoCRS)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import logging
from collections.abc import Iterable, Sequence

from odc.geo import CRS, Geometry
from typing_extensions import override

from datacube.index.abstract import AbstractProductResource
from datacube.model import Product, QueryDict, QueryField
from datacube.utils.changes import Change
from datacube.utils.json_types import JsonDict

_LOG: logging.Logger = logging.getLogger(__name__)


class ProductResource(AbstractProductResource):
    @override
    def add(self, product: Product, allow_table_lock: bool = False) -> Product | None:
        raise NotImplementedError()

    @override
    def can_update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        raise NotImplementedError()

    @override
    def update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> Product | None:
        raise NotImplementedError()

    @override
    def delete(
        self, products: Iterable[Product], allow_delete_active: bool = False
    ) -> Sequence[Product]:
        raise NotImplementedError()

    @override
    def get_unsafe(self, id_: int) -> Product:
        raise KeyError(id_)

    @override
    def get_by_name_unsafe(self, name: str) -> Product:
        raise KeyError(name)

    @override
    def search_robust(self, **query: QueryField) -> Iterable[tuple[Product, QueryDict]]:
        return []

    @override
    def search_by_metadata(self, metadata: JsonDict) -> Iterable[Product]:
        return []

    @override
    def get_all(self) -> Iterable[Product]:
        return []

    @override
    def temporal_extent(
        self, product: str | Product
    ) -> tuple[datetime.datetime, datetime.datetime]:
        raise KeyError(str(product))

    @override
    def spatial_extent(
        self, product: str | Product, crs: CRS = CRS("EPSG:4326")
    ) -> Geometry | None:
        raise KeyError(str(product))

    @override
    def most_recent_change(self, product: str | Product) -> datetime.datetime | None:
        raise KeyError(str(product))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
User Interface Utilities
"""

from datacube.utils import read_documents

from .common import get_metadata_path
from .expression import parse_expressions

__all__ = [
    "get_metadata_path",
    "parse_expressions",
    "read_documents",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

import copy

import pytest
import yaml
from sqlalchemy import text

import datacube.index.postgis.index
import datacube.index.postgres.index
from datacube.drivers.postgis._fields import (
    NumericRangeDocField as PgsNumericRangeDocField,
)
from datacube.drivers.postgis._fields import PgField as PgsPgField
from datacube.drivers.postgres._fields import (
    NumericRangeDocField as PgrNumericRangeDocField,
)
from datacube.drivers.postgres._fields import PgField as PgrPgField
from datacube.index import Index
from datacube.index.abstract import default_metadata_type_docs
from datacube.model import Dataset, DatasetType, MetadataType, Not, Product, Range
from datacube.testutils import sanitise_doc, suppress_deprecations
from datacube.utils import changes
from datacube.utils.documents import documents_equal

_DATASET_METADATA = {
    "id": "f7018d80-8807-11e5-aeaa-1040f381a756",
    "instrument": {"name": "TM"},
    "platform": {"code": "LANDSAT_5", "label": "Landsat 5"},
    "size_bytes": 4550,
    "product_type": "NBAR",
    "bands": {
        "1": {
            "type": "reflective",
            "cell_size": 25.0,
            "path": "product/LS8_OLITIRS_NBAR_P54_GALPGS01-002_112_079_20140126_B1.tif",
            "label": "Coastal Aerosol",
            "number": "1",
        },
        "2": {
            "type": "reflective",
            "cell_size": 25.0,
            "path": "product/LS8_OLITIRS_NBAR_P54_GALPGS01-002_112_079_20140126_B2.tif",
            "label": "Visible Blue",
            "number": "2",
        },
        "3": {
            "type": "reflective",
            "cell_size": 25.0,
            "path": "product/LS8_OLITIRS_NBAR_P54_GALPGS01-002_112_079_20140126_B3.tif",
            "label": "Visible Green",
            "number": "3",
        },
    },
}


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_metadata_indexes_views_exist(
    index: Index, default_metadata_type: MetadataType
) -> None:
    # Metadata indexes should no longer exist.
    assert not _object_exists(index, "dix_eo_platform")

    # Ensure view was created (following naming conventions)
    assert _object_exists(index, "dv_eo_dataset")


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_indexes_views_exist(index: Index, ls5_telem_type: DatasetType) -> None:
    assert ls5_telem_type.name == "ls5_telem_test"

    # Ensure field indexes were created for the dataset type (following the naming conventions):
    assert _object_exists(index, "dix_ls5_telem_test_orbit")

    # Ensure it does not create a 'platform' index, because that's a fixed field
    # (ie. identical in every dataset of the type)
    assert not _object_exists(index, "dix_ls5_telem_test_platform")

    # Ensure view was created (following naming conventions)
    assert _object_exists(index, "dv_ls5_telem_test_dataset")

    # Ensure view was created (following naming conventions)
    assert not _object_exists(index, "dix_ls5_telem_test_gsi"), (
        "indexed=false field gsi shouldn't have an index"
    )


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_dataset_composite_indexes_exist(index: Index, ls5_telem_type) -> None:
    # This type has fields named lat/lon/time, so composite indexes should now exist for them:
    # (following the naming conventions)
    assert _object_exists(index, "dix_ls5_telem_test_sat_path_sat_row_time")

    # But no individual field indexes for these
    assert not _object_exists(index, "dix_ls5_telem_test_sat_path")
    assert not _object_exists(index, "dix_ls5_telem_test_sat_row")
    assert not _object_exists(index, "dix_ls5_telem_test_time")


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_field_expression_unchanged(
    default_metadata_type: MetadataType, telemetry_metadata_type: MetadataType
) -> None:
    # We're checking for accidental changes here in our field-to-SQL code

    # If we started outputting a different expression they would quietly no longer match the expression
    # indexes that exist in our DBs.

    # The time field on the default 'eo' metadata type.
    field = default_metadata_type.dataset_fields["time"]
    assert isinstance(field, PgrPgField | PgsPgField)
    assert field.sql_expression == (
        "tstzrange("
        "least("
        "agdc.common_timestamp(agdc.dataset.metadata #>> '{extent, from_dt}'), "
        "agdc.common_timestamp(agdc.dataset.metadata #>> '{extent, center_dt}')"
        "), greatest("
        "agdc.common_timestamp(agdc.dataset.metadata #>> '{extent, to_dt}'), "
        "agdc.common_timestamp(agdc.dataset.metadata #>> '{extent, center_dt}')"
        "), '[]')"
    )

    field = default_metadata_type.dataset_fields["lat"]
    assert isinstance(field, PgrPgField | PgsPgField)
    assert field.sql_expression == (
        "agdc.float8range("
        "least("
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ur, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, lr, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ul, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ll, lat}' AS DOUBLE PRECISION)), "
        "greatest("
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ur, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, lr, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ul, lat}' AS DOUBLE PRECISION), "
        "CAST(agdc.dataset.metadata #>> '{extent, coord, ll, lat}' AS DOUBLE PRECISION)"
        "), '[]')"
    )

    # A single string value
    field = default_metadata_type.dataset_fields["platform"]
    assert isinstance(field, PgrPgField | PgsPgField)
    assert field.sql_expression == ("agdc.dataset.metadata #>> '{platform, code}'")

    # A single integer value
    field = telemetry_metadata_type.dataset_fields["orbit"]
    assert isinstance(field, PgrPgField | PgsPgField)
    assert field.sql_expression == (
        "CAST(agdc.dataset.metadata #>> '{acquisition, platform_orbit}' AS INTEGER)"
    )


def _object_exists(index: Index, index_name) -> bool:
    assert isinstance(
        index, datacube.index.postgres.index.Index | datacube.index.postgis.index.Index
    )
    schema_name = "odc" if index._db.driver_name == "postgis" else "agdc"
    with index._active_connection() as connection:
        val = connection._connection.execute(
            text(f"SELECT to_regclass('{schema_name}.{index_name}')")
        ).scalar()
    return val in (index_name, f"{schema_name}.{index_name}")


def test_idempotent_add_dataset_type(
    index: Index, ls8_eo3_product, extended_eo3_product_doc
) -> None:
    assert index.products.get_by_name(ls8_eo3_product.name) is not None

    # Re-add should have no effect, because it's equal to the current one.
    index.products.add_document(extended_eo3_product_doc)

    # But if we add the same type with differing properties we should get an error:
    different_product = copy.deepcopy(extended_eo3_product_doc)
    different_product["metadata"]["properties"]["eo:platform"] = "spamsat-13-and-a-half"
    with pytest.raises(changes.DocumentMismatchError):
        index.products.add_document(different_product)

        # TODO: Support for adding/changing search fields?


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_update_dataset(
    index: Index, ls5_telem_doc, example_ls5_nbar_metadata_doc
) -> None:
    ls5_telem_type = index.products.add_document(ls5_telem_doc)
    assert ls5_telem_type

    example_ls5_nbar_metadata_doc["lineage"]["source_datasets"] = {}
    dataset = Dataset(
        ls5_telem_type,
        example_ls5_nbar_metadata_doc,
        uri="file:///test/doc.yaml",
        sources={},
    )
    dataset = index.datasets.add(dataset)
    assert dataset

    # update with the same doc should do nothing
    index.datasets.update(dataset)
    updated = index.datasets.get(dataset.id)
    assert updated is not None
    assert updated.local_uri == "file:///test/doc.yaml"
    assert updated.uri == "file:///test/doc.yaml"

    # update location
    d = index.datasets.get(dataset.id)
    assert d is not None
    assert d.local_uri == "file:///test/doc.yaml"
    update = Dataset(
        ls5_telem_type,
        example_ls5_nbar_metadata_doc,
        uri="file:///test/doc2.yaml",
        sources={},
    )
    index.datasets.update(update)
    updated = index.datasets.get(dataset.id)
    assert updated is not None
    # New locations are no longer appended on update.
    # They may be indexing the same dataset from a different location: we don't want to remove the original location.
    # Returns the most recently added
    assert updated.local_uri == "file:///test/doc2.yaml"
    assert updated._uris == ["file:///test/doc2.yaml"]

    # adding more metadata should always be allowed
    doc = copy.deepcopy(updated.metadata_doc)
    doc["test1"] = {"some": "thing"}
    update = Dataset(ls5_telem_type, doc, uri=updated.uri)
    index.datasets.update(update)
    updated = index.datasets.get(dataset.id)
    assert updated is not None
    assert updated.metadata_doc["test1"] == {"some": "thing"}
    assert updated.local_uri == "file:///test/doc2.yaml"
    assert len(updated._uris) == 1

    # adding more metadata and changing location
    doc = copy.deepcopy(updated.metadata_doc)
    doc["test2"] = {"some": "other thing"}
    with suppress_deprecations():
        update = Dataset(  # Test deprecated functionality
            ls5_telem_type,
            doc,
            uris=["file:///test/doc3.yaml", "file:///test/doc3a.yaml"],
        )
        index.datasets.update(update)
        updated = index.datasets.get(dataset.id)
        assert updated is not None
        assert updated.metadata_doc["test1"] == {"some": "thing"}
        assert updated.metadata_doc["test2"] == {"some": "other thing"}
        assert updated.local_uri == "file:///test/doc3.yaml"
        assert len(updated.uris) == 3  # Test deprecated property

        # changing existing metadata fields isn't allowed by default
        doc = copy.deepcopy(updated.metadata_doc)
        doc["product_type"] = "foobar"
        update = Dataset(ls5_telem_type, doc, uri="file:///test/doc4.yaml")
        with pytest.raises(ValueError):
            index.datasets.update(update)
        updated = index.datasets.get(dataset.id)
        assert updated is not None
        assert updated.metadata_doc["test1"] == {"some": "thing"}
        assert updated.metadata_doc["test2"] == {"some": "other thing"}
        assert updated.metadata_doc["product_type"] == "nbar"
        assert updated.local_uri == "file:///test/doc3.yaml"
        assert len(updated._uris) == 3

    # allowed changes go through
    doc = copy.deepcopy(updated.metadata_doc)
    doc["product_type"] = "foobar"


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_update_product_type(
    index: Index, ls5_telem_type: Product, ls5_telem_doc, ga_metadata_type_doc
) -> None:
    assert index.products.get_by_name(ls5_telem_type.name) is not None

    # Update with a new description
    ls5_telem_doc["description"] = "New description"
    index.products.update_document(ls5_telem_doc)
    p = index.products.get_by_name(ls5_telem_type.name)
    assert p is not None
    # Ensure was updated
    assert p.definition["description"] == "New description"

    # Remove some match rules (looser rules -- that match more datasets -- should be allowed)
    assert "format" in ls5_telem_doc["metadata"]
    del ls5_telem_doc["metadata"]["format"]["name"]
    del ls5_telem_doc["metadata"]["format"]
    index.products.update_document(ls5_telem_doc)
    # Ensure was updated
    updated_type = index.products.get_by_name(ls5_telem_type.name)
    assert updated_type is not None
    assert updated_type.definition["metadata"] == ls5_telem_doc["metadata"]

    # Specifying metadata type definition (rather than name) should be allowed
    full_doc = copy.deepcopy(ls5_telem_doc)
    full_doc["metadata_type"] = ga_metadata_type_doc
    index.products.update_document(full_doc)

    # Remove fixed field, forcing a new index to be created (as datasets can now differ for the field).
    assert not _object_exists(index, "dix_ls5_telem_test_product_type")
    del ls5_telem_doc["metadata"]["product_type"]
    index.products.update_document(ls5_telem_doc)
    # Ensure was updated
    assert _object_exists(index, "dix_ls5_telem_test_product_type")
    updated_type = index.products.get_by_name(ls5_telem_type.name)
    assert updated_type is not None
    assert updated_type.definition["metadata"] == ls5_telem_doc["metadata"]

    # But if we make metadata more restrictive we get an error:
    different_telemetry_type = copy.deepcopy(ls5_telem_doc)
    assert "ga_label" not in different_telemetry_type["metadata"]
    different_telemetry_type["metadata"]["ga_label"] = "something"
    with pytest.raises(ValueError):
        index.products.update_document(different_telemetry_type)
    # Check was not updated.
    updated_type = index.products.get_by_name(ls5_telem_type.name)
    assert updated_type is not None
    assert "ga_label" not in updated_type.definition["metadata"]

    # But works when unsafe updates are allowed.
    index.products.update_document(different_telemetry_type, allow_unsafe_updates=True)
    updated_type = index.products.get_by_name(ls5_telem_type.name)
    assert updated_type is not None
    assert updated_type.definition["metadata"]["ga_label"] == "something"


def test_product_update_cli(
    index: Index,
    clirunner,
    ls8_eo3_product: Product,
    extended_eo3_product_doc: dict,
    extended_eo3_metadata_type: MetadataType,
    tmpdir,
) -> None:
    """
    Test updating products via cli
    """

    def run_update_product(
        file_path, expect_success: bool = True, allow_unsafe: bool = False
    ):
        return clirunner(
            ["product", "update", str(file_path)]
            + (["--allow-unsafe"] if allow_unsafe else []),
            catch_exceptions=False,
            expect_success=expect_success,
        )

    def get_current(index: Index, product_doc: dict) -> str | float | dict | list:
        # It's calling out to a separate instance to update the product (through the cli),
        # so we need to clear our local index object's cache to get the updated one.
        index.products.get_by_name_unsafe.cache_clear()  # type: ignore[attr-defined]
        p = index.products.get_by_name(product_doc["name"])
        assert p is not None
        return sanitise_doc(p.definition)

    # Update an unchanged file, should be unchanged.
    file_path = tmpdir.join("unmodified-product.yaml")
    file_path.write(_to_yaml(extended_eo3_product_doc))
    result = run_update_product(file_path)
    assert str(f'Updated "{extended_eo3_product_doc["name"]}"') in result.output
    fresh = get_current(index, extended_eo3_product_doc)
    assert documents_equal(fresh, extended_eo3_product_doc)
    assert result.exit_code == 0

    # Try to add an unknown property: this should be forbidden by validation of dataset-type-schema.yaml
    modified_doc = copy.deepcopy(extended_eo3_product_doc)
    modified_doc["newly_added_property"] = {}
    file_path = tmpdir.join("invalid-product.yaml")
    file_path.write(_to_yaml(modified_doc))
    result = run_update_product(file_path, expect_success=False)

    # The error message differs between jsonschema versions, but should always mention the invalid property name.
    assert "newly_added_property" in result.output
    # Return error code for failure!
    assert result.exit_code == 1
    fresh = get_current(index, extended_eo3_product_doc)
    assert documents_equal(fresh, extended_eo3_product_doc)

    # Use of a numeric key in the document
    # (This has thrown errors in the past. all dict keys are strings after json conversion, but some old docs use
    # numbers as keys in yaml)
    modified_doc = copy.deepcopy(extended_eo3_product_doc)
    modified_doc["metadata"][42] = "hello"
    file_path = tmpdir.join("unsafe-change-to-product.yaml")
    file_path.write(_to_yaml(modified_doc))
    result = run_update_product(file_path, expect_success=False)
    assert "Unsafe change in metadata.42 from missing to 'hello'" in result.output
    # Return error code for failure!
    assert result.exit_code == 1
    # Unchanged
    fresh = get_current(index, extended_eo3_product_doc)
    assert documents_equal(fresh, extended_eo3_product_doc)

    # But if we set allow-unsafe==True, this one will work.
    result = run_update_product(file_path, allow_unsafe=True)
    assert "Unsafe change in metadata.42 from missing to 'hello'" in result.output
    assert result.exit_code == 0
    # Has changed, and our key is now a string (json only allows string keys)
    modified_doc = copy.deepcopy(extended_eo3_product_doc)
    modified_doc["metadata"]["42"] = "hello"
    fresh = get_current(index, extended_eo3_product_doc)
    assert documents_equal(fresh, modified_doc)


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_product_delete(index: Index, ls8_eo3_product: Product) -> None:
    # test that postgres dynamic indexes and views are deleted
    assert index.products.get_by_name(ls8_eo3_product.name) is not None
    assert _object_exists(index, "dix_ga_ls8c_ard_3_region_code")
    assert _object_exists(index, "dv_ga_ls8c_ard_3_dataset")

    index.products.delete([ls8_eo3_product])

    index.products.get_by_name_unsafe.cache_clear()  # type: ignore[attr-defined]
    assert index.products.get_by_name(ls8_eo3_product.name) is None
    assert not _object_exists(index, "dix_ga_ls8c_ard_3_region_code")
    assert not _object_exists(index, "dv_ga_ls8c_ard_3_dataset")


def test_product_delete_cli(
    index: Index,
    clirunner,
    ls8_eo3_product: Product,
    ls8_eo3_dataset,
    ls8_eo3_dataset2,
    wo_eo3_product,
) -> None:
    from pathlib import Path

    TESTDIR = Path(__file__).parent.parent / "data" / "eo3"  # noqa: N806
    # product with some archived and some active datasets
    clirunner(["dataset", "archive", "c21648b1-a6fa-4de0-9dc3-9c445d8b295a"])

    runner = clirunner(
        ["product", "delete", "ga_ls8c_ard_3", "ga_ls_wo_3"], verbose_flag=False
    )
    assert "1 out of 2 products successfully deleted" in runner.output
    assert "ga_ls_wo_3 could not be deleted" not in runner.output
    assert "ga_ls8c_ard_3 could not be deleted" in runner.output
    assert runner.exit_code == 0

    index.products.get_by_name_unsafe.cache_clear()  # type: ignore[attr-defined]
    assert index.products.get_by_name("ga_ls8c_ard_3") is not None
    assert index.products.get_by_name("ga_ls_wo_3") is None

    # adding back product should cause error since it still exists
    add = clirunner(["product", "add", str(TESTDIR / "ard_ls8.odc-product.yaml")])
    assert "is already in the database" in add.output

    # ensure deletion involving active datasets works with force
    runner = clirunner(
        ["product", "delete", "ga_ls8c_ard_3", "--force"], verbose_flag=False
    )
    assert "Completed product deletion" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "archive", "4a30d008-4e82-4d67-99af-28bc1629f766"],
        verbose_flag=False,
        expect_success=False,
    )
    assert "No dataset found with id" in runner.output

    index.products.get_by_name_unsafe.cache_clear()  # type: ignore[attr-defined]
    assert index.products.get_by_name("ga_ls8c_ard_3") is None

    # should be able to add product back now
    add = clirunner(["product", "add", str(TESTDIR / "ard_ls8.odc-product.yaml")])
    assert "is already in the database" not in add.output


def _to_yaml(ls5_telem_doc):
    # Need to explicitly allow unicode in Py2
    return yaml.safe_dump(ls5_telem_doc, allow_unicode=True)


def test_update_metadata_type(
    index: Index, default_metadata_type: MetadataType
) -> None:
    mt_doc = next(
        d
        for d in default_metadata_type_docs()
        if d["name"] == default_metadata_type.name
    )

    assert index.metadata_types.get_by_name(mt_doc["name"]) is not None

    # Update with no changes should work.
    index.metadata_types.update_document(mt_doc)

    # Add search field
    mt_doc["dataset"]["search_fields"]["testfield"] = {
        "description": "Field added for testing",
        "offset": ["properties", "test"],
    }

    # TODO: Able to remove fields?
    # Indexes will be difficult to handle, as dropping them may affect other users. But leaving them there may
    # lead to issues if a different field is created with the same name.

    index.metadata_types.update_document(mt_doc)
    # Ensure was updated
    updated_type = index.metadata_types.get_by_name(mt_doc["name"])
    assert updated_type is not None
    assert "testfield" in updated_type.dataset_fields

    # But if we change an existing field type we get an error:
    different_mt_doc = copy.deepcopy(mt_doc)
    different_mt_doc["dataset"]["search_fields"]["time"]["type"] = "numeric-range"
    with pytest.raises(ValueError):
        index.metadata_types.update_document(different_mt_doc)

    # But works when unsafe updates are allowed.
    index.metadata_types.update_document(different_mt_doc, allow_unsafe_updates=True)
    updated_type = index.metadata_types.get_by_name(mt_doc["name"])
    assert updated_type is not None
    assert isinstance(
        updated_type.dataset_fields["time"],
        PgrNumericRangeDocField | PgsNumericRangeDocField,
    )


def test_filter_types_by_fields(index: Index, wo_eo3_product) -> None:
    res = list(
        index.metadata_types.get_with_fields(["platform", "instrument", "region_code"])
    )
    assert res == [wo_eo3_product.metadata_type]

    res = list(
        index.metadata_types.get_with_fields(
            ["platform", "instrument", "region_code", "favorite_icecream"]
        )
    )
    assert len(res) == 0

    res = list(
        index.products.get_with_fields(["platform", "instrument", "region_code"])
    )
    assert res == [wo_eo3_product]

    res = list(
        index.products.get_with_fields(
            ["platform", "instrument", "region_code", "favorite_icecream"]
        )
    )
    assert len(res) == 0


def test_filter_products_by_types(index: Index, wo_eo3_product) -> None:
    res = list(index.products.get_with_types([wo_eo3_product.metadata_type]))
    assert res == [wo_eo3_product]


def test_filter_types_by_search(index: Index, wo_eo3_product, ls8_eo3_product) -> None:
    # No arguments, return all.
    res = set(index.products.search())
    assert res == {ls8_eo3_product, wo_eo3_product}

    # Matching fields
    res = list(index.products.search(product_family="wo", product="ga_ls_wo_3"))
    assert res == [wo_eo3_product]

    # Matching fields and non-available fields
    res = list(
        index.products.search(
            product_family="wo",
            product="ga_ls_wo_3",
            lat=Range(142.015625, 142.015625),
            lon=Range(-12.046875, -12.046875),
        )
    )
    assert res == []

    # Matching fields and available fields
    [(res, q)] = list(
        index.products.search_robust(
            product_family="wo",
            product="ga_ls_wo_3",
            cloud_cover=Range(0.015625, 0.2015625),
            dataset_maturity="final",
        )
    )
    assert res == wo_eo3_product
    assert "cloud_cover" in q
    assert "dataset_maturity" in q

    # Or expression test
    res = list(index.products.search(product_family=["wo", "spam"]))
    assert res == [wo_eo3_product]

    # Not expression test
    res = list(index.products.search(product_family=Not("wo")))
    assert res == [ls8_eo3_product]

    # Mismatching fields
    res = list(index.products.search(product_family="spam"))
    assert res == []


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_update_metadata_type_doc(index: Index, ls5_telem_type) -> None:
    type_doc = copy.deepcopy(ls5_telem_type.metadata_type.definition)
    type_doc["dataset"]["search_fields"]["test_indexed"] = {
        "description": "indexed test field",
        "offset": ["test", "indexed"],
    }
    type_doc["dataset"]["search_fields"]["test_not_indexed"] = {
        "description": "not indexed test field",
        "offset": ["test", "not", "indexed"],
        "indexed": False,
    }

    index.metadata_types.update_document(type_doc)

    assert ls5_telem_type.name == "ls5_telem_test"
    assert _object_exists(index, "dix_ls5_telem_test_test_indexed")
    assert not _object_exists(index, "dix_ls5_telem_test_test_not_indexed")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np
import pytest
import yaml
from xarray import DataArray, Dataset

from datacube.utils.masking import (
    create_mask_value,
    describe_variable_flags,
    list_flag_names,
    mask_invalid_data,
    mask_to_dict,
    valid_data_mask,
)


@pytest.fixture
def simple_var():
    flags = SimpleVariableWithFlagsDef().flags_definition
    return DataArray(
        np.zeros((2, 3)),
        dims=("y", "x"),
        name="simple_var",
        attrs={"flags_definition": flags},
    )


def test_list_flag_names(simple_var) -> None:
    flags = list_flag_names(simple_var)
    for flag_name in simple_var.flags_definition:
        assert flag_name in flags

    with pytest.raises(ValueError):
        list_flag_names(([], {}))


def test_create_mask_value(simple_var) -> None:
    bits_def = simple_var.flags_definition

    assert create_mask_value(bits_def, contiguous=True) == (256, 256)
    assert create_mask_value(bits_def, contiguous=False) == (256, 0)
    assert create_mask_value(bits_def, contiguous=True, land_sea="land") == (768, 768)
    assert create_mask_value(bits_def, contiguous=False, land_sea="land") == (768, 512)


def test_create_multi_mask_value() -> None:
    multi_var = VariableWithMultiBitFlags()
    multi_flags_def = multi_var.flags_definition

    assert create_mask_value(multi_flags_def, filled=True) == (1, 1)
    assert create_mask_value(multi_flags_def, water_confidence="water") == (
        0b011000,
        0b011000,
    )

    assert create_mask_value(
        multi_flags_def, water_confidence="water", filled=True
    ) == (0b011001, 0b011001)
    assert create_mask_value(multi_flags_def, water_confidence="not_determined") == (
        0b011000,
        0b0,
    )
    assert create_mask_value(multi_flags_def, water_confidence="no_water") == (
        0b11000,
        0b01000,
    )
    assert create_mask_value(multi_flags_def, veg_confidence="maybe_veg") == (
        0b110000000,
        0b100000000,
    )
    assert create_mask_value(
        multi_flags_def, veg_confidence="maybe_veg", water_confidence="water"
    ) == (0b110011000, 0b100011000)
    assert create_mask_value(
        multi_flags_def,
        veg_confidence="maybe_veg",
        water_confidence="water",
        filled=True,
    ) == (0b110011001, 0b100011001)

    assert create_mask_value(multi_flags_def, water_confidence="maybe_water") == (
        0b011000,
        0b10000,
    )

    with pytest.raises(ValueError):
        create_mask_value(multi_flags_def, this_flag_doesnot_exist=9)

    with pytest.raises(ValueError):
        create_mask_value(multi_flags_def, water_confidence="invalid enum value")


def test_ga_good_pixel(simple_var) -> None:
    bits_def = simple_var.flags_definition

    assert create_mask_value(bits_def, ga_good_pixel=True) == (16383, 16383)


def test_describe_flags(simple_var) -> None:
    describe_variable_flags(simple_var)
    describe_variable_flags(simple_var, with_pandas=False)

    describe_variable_flags(simple_var.to_dataset())
    describe_variable_flags(simple_var.to_dataset(), with_pandas=False)


class SimpleVariableWithFlagsDef:
    bits_def_yaml = """
        cloud_shadow_fmask:
          bits: 13
          description: Cloud Shadow (Fmask)
          values:
            0: cloud_shadow
            1: no_cloud_shadow
        cloud_shadow_acca:
          bits: 12
          description: Cloud Shadow (ACCA)
          values:
            0: cloud_shadow
            1: no_cloud_shadow
        cloud_fmask:
          bits: 11
          description: Cloud (Fmask)
          values:
            0: cloud
            1: no_cloud
        cloud_acca:
          bits: 10
          description: Cloud Shadow (ACCA)
          values:
            0: cloud
            1: no_cloud
        land_sea:
          bits: 9
          description: Land or Sea
          values:
            0: sea
            1: land
        contiguous:
          bits: 8
          description: All bands for this pixel contain non-null values
          values:
            0: false
            1: true
        swir2_saturated:
          bits: 7
          description: SWIR2 is saturated
          values:
            0: true
            1: false

        swir2_saturated:
          bits: 6
          description: SWIR2 band is saturated
          values:
            0: true
            1: false
        swir1_saturated:
          bits: 4
          description: SWIR1 band is saturated
          values:
            0: true
            1: false
        nir_saturated:
          bits: 3
          description: NIR band is saturated
          values:
            0: true
            1: false

        red_saturated:
          bits: 2
          description: Red band is saturated
          values:
            0: true
            1: false

        green_saturated:
          bits: 1
          description: Green band is saturated
          values:
            0: true
            1: false

        blue_saturated:
          bits: 0
          description: Blue band is saturated
          values:
            0: true
            1: false

        ga_good_pixel:
          bits: [13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
          description: Best Quality Pixel
          values:
            16383: true
        """
    flags_definition = yaml.safe_load(bits_def_yaml)


class VariableWithMultiBitFlags:
    bits_def_yaml = """
        cloud_confidence:
          bits: [13, 14]
          description: Cloud Confidence
          values:
            0: not_determined
            1: no_cloud
            2: maybe_cloud
            3: cloud
        cirrus_confidence:
          bits: [11, 12]
          description: Cirrus Confidence
          values:
            0: not_determined
            1: no_cirrus
            2: maybe_cirrus
            3: cirrus

        snowice_confidence:
          bits: [9, 10]
          description: Snow/Ice Confidence
          values:
            0: not_determined
            1: no_snowice
            2: maybe_snowice
            3: snowice

        veg_confidence:
          bits: [7, 8]
          description: Vegetation Confidence
          values:
            0: not_determined
            1: no_veg
            2: maybe_veg
            3: veg



        water_confidence:
          bits: [3, 4]
          description: Water Confidence
          values:
            0: not_determined
            1: no_water
            2: maybe_water
            3: water

        terrain_occluded:
          bits: 2
          description: Terrain occluded
          values:
            0: False
            1: True

        frame_dropped:
          bits: 1
          description: Frame dropped
          values:
            0: False
            1: True

        filled:
          bits: 0
          description: Filled
          values:
            0: False
            1: True
    """
    flags_definition = yaml.safe_load(bits_def_yaml)


bits_def_json = {
    "blue_saturated": {
        "bits": 0,
        "description": "Blue band is saturated",
        "values": {"0": True, "1": False},
    },
    "cloud_acca": {
        "bits": 10,
        "description": "Cloud Shadow (ACCA)",
        "values": {"0": "cloud", "1": "no_cloud"},
    },
    "cloud_fmask": {
        "bits": 11,
        "description": "Cloud (Fmask)",
        "values": {"0": "cloud", "1": "no_cloud"},
    },
    "cloud_shadow_acca": {
        "bits": 12,
        "description": "Cloud Shadow (ACCA)",
        "values": {"0": "cloud_shadow", "1": "no_cloud_shadow"},
    },
    "cloud_shadow_fmask": {
        "bits": 13,
        "description": "Cloud Shadow (Fmask)",
        "values": {"0": "cloud_shadow", "1": "no_cloud_shadow"},
    },
    "contiguous": {
        "bits": 8,
        "description": "All bands for this pixel contain non-null values",
        "values": {"0": False, "1": True},
    },
    "ga_good_pixel": {
        "bits": [13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
        "description": "Best Quality Pixel",
        "values": {"16383": True},
    },
    "green_saturated": {
        "bits": 1,
        "description": "Green band is saturated",
        "values": {"0": True, "1": False},
    },
    "land_sea": {
        "bits": 9,
        "description": "Land or Sea",
        "values": {"0": "sea", "1": "land"},
    },
    "nir_saturated": {
        "bits": 3,
        "description": "NIR band is saturated",
        "values": {"0": True, "1": False},
    },
    "red_saturated": {
        "bits": 2,
        "description": "Red band is saturated",
        "values": {"0": True, "1": False},
    },
    "swir1_saturated": {
        "bits": 4,
        "description": "SWIR1 band is saturated",
        "values": {"0": True, "1": False},
    },
    "swir2_saturated": {
        "bits": 7,
        "description": "SWIR2 band is saturated",
        "values": {"0": True, "1": False},
    },
    "tir_saturated": {
        "bits": 5,
        "description": "Thermal Infrared band is saturated",
        "values": {"0": True, "1": False},
    },
}


@pytest.mark.parametrize(
    "bits_def", [SimpleVariableWithFlagsDef.flags_definition, bits_def_json]
)
def test_simple_mask_to_dict(bits_def) -> None:
    # All 0. Contiguous should be False, Saturated bits should be true
    test_dict = mask_to_dict(bits_def, 0)
    assert not test_dict["contiguous"]
    assert test_dict["blue_saturated"]

    # Only contiguous (bit 8) set
    test_dict = mask_to_dict(bits_def, 256)
    assert test_dict["contiguous"]
    assert test_dict["blue_saturated"]
    assert test_dict["land_sea"] == "sea"

    # Contiguous and land_sea bits set to 1. (bits 7 and 8)
    test_dict = mask_to_dict(bits_def, 768)
    assert test_dict["contiguous"]
    assert test_dict["land_sea"] == "land"
    assert test_dict["blue_saturated"]

    # All GA PQ bits set to 1
    test_dict = mask_to_dict(bits_def, 16383)
    assert test_dict["ga_good_pixel"]
    assert test_dict["contiguous"]
    assert test_dict["land_sea"] == "land"
    assert not test_dict["blue_saturated"]


def test_mask_valid_data() -> None:
    test_attrs = {
        "one": 1,
        "nodata": -999,
    }

    expected_data_array = DataArray(
        np.array(
            [[1.0, np.nan, np.nan], [2, 3, np.nan], [np.nan, np.nan, np.nan]],
            dtype="float",
        ),
        attrs=test_attrs,
        name="var_one",
    )

    data_array = DataArray(
        [[1, -999, -999], [2, 3, -999], [-999, -999, -999]], attrs=test_attrs
    )
    dataset = Dataset(
        data_vars={"var_one": data_array}, attrs={"ds_attr": "still here"}
    )

    # Make sure test is actually changing something
    assert not data_array.equals(expected_data_array)

    output_ds = mask_invalid_data(dataset, keep_attrs=True)
    assert output_ds.attrs["ds_attr"] == "still here"
    assert output_ds.data_vars["var_one"].equals(expected_data_array)
    assert output_ds.data_vars["var_one"].attrs["one"] == 1

    output_da = mask_invalid_data(data_array, keep_attrs=True)
    assert output_da.equals(expected_data_array)
    assert output_da.attrs["one"] == 1

    missing_nodata = data_array.copy()
    del missing_nodata.attrs["nodata"]
    assert not hasattr(missing_nodata, "nodata")
    np.testing.assert_array_equal(missing_nodata, mask_invalid_data(missing_nodata))

    with pytest.raises(TypeError):
        mask_invalid_data({})


def test_valid_data_mask() -> None:
    attrs = {
        "nodata": -999,
    }

    expected_data_array = DataArray(
        np.array(
            [[True, False, False], [True, True, False], [False, False, False]],
            dtype="bool",
        )
    )

    data_array = DataArray(
        [[1, -999, -999], [2, 3, -999], [-999, -999, -999]], attrs=attrs
    )
    dataset = Dataset(data_vars={"var_one": data_array})

    output_ds = valid_data_mask(dataset)
    assert output_ds.data_vars["var_one"].equals(expected_data_array)

    output_da = valid_data_mask(data_array)
    assert output_da.equals(expected_data_array)

    expected_data_array = DataArray(
        np.array(
            [[True, True, True], [True, True, True], [True, True, True]], dtype="bool"
        )
    )
    data_array = DataArray([[1, -999, -999], [2, 3, -999], [-999, -999, -999]])
    dataset = Dataset(data_vars={"var_one": data_array})

    output_ds = valid_data_mask(dataset)
    assert output_ds.data_vars["var_one"].equals(expected_data_array)

    output_da = valid_data_mask(data_array)
    assert output_da.equals(expected_data_array)

    expected_data_array = DataArray(
        np.array(
            [[True, False, False], [True, True, False], [False, False, False]],
            dtype="bool",
        )
    )

    data_array = DataArray(
        [[1, -999, -999], [2, 3, -999], [-999, -999, float("nan")]], attrs=attrs
    )
    dataset = Dataset(data_vars={"var_one": data_array})

    output_ds = valid_data_mask(dataset)
    assert output_ds.data_vars["var_one"].equals(expected_data_array)

    output_da = valid_data_mask(data_array)
    assert output_da.equals(expected_data_array)

    expected_data_array = DataArray(
        np.array(
            [[True, True, True], [True, True, True], [True, True, False]], dtype="bool"
        )
    )

    data_array = DataArray([[1, -999, -999], [2, 3, -999], [-999, -999, float("nan")]])
    dataset = Dataset(data_vars={"var_one": data_array})

    output_ds = valid_data_mask(dataset)
    assert output_ds.data_vars["var_one"].equals(expected_data_array)

    output_da = valid_data_mask(data_array)
    assert output_da.equals(expected_data_array)

    with pytest.raises(TypeError):
        valid_data_mask(([], []))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

import datetime
from os import terminal_size

from dateutil.tz import UTC
from sqlalchemy.dialects.postgresql import Range as PgRange

from datacube.index.fields import Field
from datacube.scripts.search_tool import write_csv, write_pretty


class MockFile:
    def __init__(self) -> None:
        self.vals: list = []

    def write(self, s) -> None:
        self.vals.append(s)

    def flush(self) -> None:
        pass

    def getvalue(self):
        return "".join(self.vals)


def test_csv_serialise() -> None:
    m = MockFile()
    write_csv(
        m,
        {"f1": Field("f1", ""), "f2": Field("f2", "")},
        [
            {"f1": 12, "f2": PgRange(1.0, 2.0)},
            {
                "f1": datetime.datetime(2014, 7, 26, 23, 48, 0, tzinfo=UTC),
                "f2": PgRange(-1.0, 2.0),
            },
            {"f1": datetime.datetime(2014, 7, 26, 23, 48, 0), "f2": "landsat"},
        ],
    )

    assert m.getvalue() == (
        "f1,f2\r\n"
        "12,1.0 to 2.0\r\n"
        "2014-07-26T23:48:00+00:00,-1.0 to 2.0\r\n"
        "2014-07-26T23:48:00+00:00,landsat\r\n"
    )


def test_pretty_serialise() -> None:
    m = MockFile()
    write_pretty(
        m,
        {"f1": Field("f1", ""), "field 2": Field("f2", "")},
        [
            {"f1": 12, "field 2": PgRange(1.0, 2.0)},
            {
                "f1": datetime.datetime(2014, 7, 26, 23, 48, 0, tzinfo=UTC),
                "field 2": PgRange(-1.0, 2.0),
            },
            {"f1": datetime.datetime(2014, 7, 26, 23, 48, 0), "field 2": "landsat"},
        ],
        terminal_size=terminal_size((12, 12)),
    )

    assert m.getvalue() == (
        "-[ 1 ]-----\n"
        "f1      | 12\n"
        "field 2 | 1.0 to 2.0\n"
        "-[ 2 ]-----\n"
        "f1      | 2014-07-26T23:48:00+00:00\n"
        "field 2 | -1.0 to 2.0\n"
        "-[ 3 ]-----\n"
        "f1      | 2014-07-26T23:48:00+00:00\n"
        "field 2 | landsat\n"
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import warnings
from collections.abc import Callable

import numpy as np
from affine import Affine
from odc.geo import CRS
from odc.geo.geobox import GeoBox
from odc.geo.gridspec import GridSpec
from odc.geo.math import apply_affine

# from datacube.model import GridSpec

# pylint: disable=invalid-name

epsg4326 = CRS("EPSG:4326")
epsg3577 = CRS("EPSG:3577")
epsg3857 = CRS("EPSG:3857")

AlbersGS = GridSpec(crs=epsg3577, tile_shape=(4000, 4000), resolution=25)

SAMPLE_WKT_WITHOUT_AUTHORITY = """PROJCS["unnamed",
       GEOGCS["unnamed ellipse",
              DATUM["unknown",
                    SPHEROID["unnamed",6378137,0],
                    EXTENSION["PROJ4_GRIDS","@null"]],
              PRIMEM["Greenwich",0],
              UNIT["degree",0.0174532925199433]],
       PROJECTION["Mercator_2SP"],
       PARAMETER["standard_parallel_1",0],
       PARAMETER["central_meridian",0],
       PARAMETER["false_easting",0],
       PARAMETER["false_northing",0],
       UNIT["Meter",1]
]
"""


def mkA(  # noqa: N802
    rot: float = 0.0,
    scale=(1, 1),
    shear: float = 0.0,
    translation: tuple[float, float] = (0.0, 0.0),
) -> Affine:
    return (
        Affine.translation(*translation)
        * Affine.rotation(rot)
        * Affine.shear(shear)
        * Affine.scale(*scale)
    )


def xy_from_gbox(gbox: GeoBox) -> tuple[np.ndarray, np.ndarray]:
    """
    :returns: Two images with X and Y coordinates for centers of pixels
    """
    h, w = gbox.shape

    xx, yy = np.meshgrid(
        np.arange(w, dtype="float64") + 0.5, np.arange(h, dtype="float64") + 0.5
    )

    return apply_affine(gbox.transform, xx, yy)


def xy_norm(
    x: np.ndarray, y: np.ndarray, deg: float = 33.0
) -> tuple[np.ndarray, np.ndarray, Affine]:
    """
    Transform output of xy_from_geobox with a reversible linear transform. On
    output x,y are in [0,1] range. Reversible Affine transform includes
    rotation by default, this is to ensure that test images don't have
    symmetries that are aligned to X/Y axis.

    1. Rotate x,y by ``deg``
    2. Offset and scale such that values are in [0, 1] range


    :returns: (x', y', A)

    - (x, y) == A*(x', y')
    - [x|y]'.min() == 0
    - [x|y]'.max() == 1
    """

    def norm_v(v):
        vmin = v.min()
        v -= vmin
        s = 1.0 / v.max()
        v *= s

        return s, -vmin * s

    A_rot = Affine.rotation(deg)  # noqa: N806
    x, y = apply_affine(A_rot, x, y)

    sx, tx = norm_v(x)
    sy, ty = norm_v(y)

    A = Affine(sx, 0, tx, 0, sy, ty) * A_rot

    return x, y, ~A


def to_fixed_point(a, dtype: str | np.dtype | type = "uint16"):
    """
    Convert normalised ([0,1]) floating point image to integer fixed point fractional.

    Note for signed types: there is no offset, 0 -> 0, 1 -> (2**(nbits - 1) - 1).

    Reverse is provided by: ``from_fixed_point``
    """
    ii = np.iinfo(dtype)
    a = a * ii.max + 0.5
    a = np.clip(a, 0, ii.max, out=a)
    warnings.filterwarnings("error")
    try:
        b = a.astype(ii.dtype)
    except RuntimeWarning as e:
        raise TypeError(e) from None
    warnings.resetwarnings()
    if not (np.abs(a - b) < 1).all():
        raise TypeError(f"Cannot safely cast float to {dtype}")
    return b


def from_fixed_point(a):
    """
    Convert fixed point image to floating point

    This is reverse of ``to_fixed_point``
    """
    ii = np.iinfo(a.dtype)
    return a.astype("float64") * (1.0 / ii.max)


def gen_test_image_xy(
    gbox: GeoBox, dtype: str | np.dtype | type = "float32", deg: float = 33.0
) -> tuple[np.ndarray, Callable]:
    """
    Generate test image that captures pixel coordinates in pixel values.
    Useful for testing reprojections/reads.

    :param gbox: GeoBox defining pixel plane

    :dtype: data type of the image, defaults to `float32`, but it can be an
            integer type in which case normalised coordinates will be
            quantised increasing error.

    :returns: 2xWxH ndarray encoding X,Y coordinates of pixel centers in some
              normalised space, and a callable that can convert from normalised
              space back to coordinate space.
    """
    dtype = np.dtype(dtype)

    x, y = xy_from_gbox(gbox)
    x, y, A = xy_norm(x, y, deg)

    xy = np.stack([x, y])

    xy = xy.astype(dtype) if dtype.kind == "f" else to_fixed_point(xy, dtype)

    def denorm(xy=None, y=None, nodata=None):
        if xy is None:
            return A

        stacked = y is None
        x, y = xy if stacked else (xy, y)
        missing_mask = None

        if nodata is not None:
            missing_mask = (
                np.isnan(x) + np.isnan(y)
                if np.isnan(nodata)
                else (x == nodata) + (y == nodata)
            )

        if x.dtype.kind != "f":
            x = from_fixed_point(x)
            y = from_fixed_point(y)

        x, y = apply_affine(A, x, y)

        if missing_mask is not None:
            x[missing_mask] = np.nan
            y[missing_mask] = np.nan

        if stacked:
            return np.stack([x, y])
        else:
            return x, y

    return xy, denorm
